from __future__ import annotations
from math import nan
from operator import attrgetter
from pathlib import Path
from typing import (TYPE_CHECKING, Callable, Dict, Iterable, Iterator, List,
                    Optional, Tuple)

import numpy as np
from ase.units import Bohr, Ha, alpha

from gpaw.band_descriptor import BandDescriptor
from gpaw.grid_descriptor import GridDescriptor
from gpaw.kpoint import KPoint
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import broadcast_array, serial_comm
from gpaw.occupations import OccupationNumberCalculator, ParallelLayout
from gpaw.projections import Projections
from gpaw.setup import Setup
from gpaw.typing import Array1D, Array2D, Array3D, Array4D, ArrayND
from gpaw.ibz2bz import IBZ2BZMaps
from gpaw.utilities.partition import AtomPartition

if TYPE_CHECKING:
    from gpaw.calculator import GPAW  # noqa
    from gpaw.new.ase_interface import ASECalculator

_L_vlmm: List[List[np.ndarray]] = []  # see get_L_vlmm() below


class WaveFunction:
    def __init__(self,
                 eigenvalues: Array1D,
                 projections: Projections,
                 bz_index: int = None):
        self.eig_m = eigenvalues
        self.projections = projections
        self.spin_projection_mv: Optional[Array2D] = None
        self.v_mn: Optional[Array2D] = None
        self.f_m = np.empty_like(self.eig_m)
        self.f_m[:] = nan
        self.bz_index = bz_index

    def transform(self, IBZ2BZMap, K) -> 'WaveFunction':
        """Transforms PAW projections from IBZ to BZ k-point."""
        projections = IBZ2BZMap.map_projections(self.projections)
        return WaveFunction(self.eig_m.copy(), projections, K)

    def redistribute_atoms(self,
                           atom_partition: AtomPartition
                           ) -> 'WaveFunction':
        projections = self.projections.redist(atom_partition)
        return WaveFunction(self.eig_m.copy(), projections, self.bz_index)

    def add_soc(self,
                dVL_avii: Dict[int, Array3D],
                s_vss: List[Array2D],
                C_ss: Array2D) -> None:
        """Evaluate H in a basis of S_z eigenstates."""
        if self.projections.bcomm.rank > 0:
            return

        M = self.projections.nbands
        H_mm = np.zeros((M, M), complex)
        for a, dVL_vii in dVL_avii.items():
            ni = dVL_vii.shape[1]
            H_ssii = np.zeros((2, 2, ni, ni), complex)
            H_ssii[0, 0] = dVL_vii[2]
            H_ssii[0, 1] = dVL_vii[0] - 1.0j * dVL_vii[1]
            H_ssii[1, 0] = dVL_vii[0] + 1.0j * dVL_vii[1]
            H_ssii[1, 1] = -dVL_vii[2]

            # Tranform to theta, phi basis
            H_ssii = np.tensordot(C_ss, H_ssii, (0, 1))
            H_ssii = np.tensordot(C_ss.T.conj(), H_ssii, (1, 1))
            H_ssii *= Ha

            P_msi = self.projections[a]
            for s1 in range(2):
                for s2 in range(2):
                    H_ii = H_ssii[s1, s2]
                    P1_mi = P_msi[:, s1]
                    P2_mi = P_msi[:, s2]
                    H_mm += np.dot(np.dot(P1_mi.conj(), H_ii), P2_mi.T)

        domain_comm = self.projections.atom_partition.comm
        domain_comm.sum(H_mm, 0)
        if domain_comm.rank == 0:
            H_mm += np.diag(self.eig_m)
            self.eig_m, v_nm = np.linalg.eigh(H_mm)
        else:
            v_nm = np.empty((M, M), complex)

        domain_comm.broadcast(v_nm, 0)

        P_mI = self.projections.matrix.array
        P_mI[:] = v_nm.T.copy().dot(P_mI)

        sx_m = []
        sy_m = []
        sz_m = []

        v_msn = v_nm.copy().reshape((M // 2, 2, M)).T.copy()
        for v_sn in v_msn:
            sx_m.append(np.trace(v_sn.T.conj().dot(s_vss[0]).dot(v_sn)))
            sy_m.append(np.trace(v_sn.T.conj().dot(s_vss[1]).dot(v_sn)))
            sz_m.append(np.trace(v_sn.T.conj().dot(s_vss[2]).dot(v_sn)))

        self.spin_projection_mv = np.array([sx_m, sy_m, sz_m]).real.T.copy()
        self.v_mn = v_nm.T

    def wavefunctions(self, calc, periodic=True):
        kd = calc.wfs.kd
        assert kd.nibzkpts == kd.nbzkpts

        # For spinors the number of bands is doubled and a
        # spin dimension is added
        Ns = calc.wfs.nspins
        Nm, Nn = self.v_mn.shape

        if calc.wfs.collinear:
            u_snR = [[calc.wfs.get_wave_function_array(n, self.bz_index, s,
                                                       periodic=periodic)
                      for n in range(Nn // 2)]
                     for s in range(Ns)]
            u_msR = np.empty((Nm, 2) + u_snR[0][0].shape, complex)
            np.einsum('mn, nabc -> mabc', self.v_mn[:, ::2], u_snR[0],
                      out=u_msR[:, 0])
            np.einsum('mn, nabc -> mabc', self.v_mn[:, 1::2], u_snR[-1],
                      out=u_msR[:, 1])
        else:
            u_nsR = np.array(
                [calc.wfs.get_wave_function_array(n, self.bz_index, 0,
                                                  periodic=periodic)
                 for n in range(Nn)])
            u_msR = np.einsum('mn, nsxyz -> msxyz',
                              self.v_mn, u_nsR)

        return u_msR

    @property
    def P_amj(self):
        M = self.projections.nbands
        return {
            a: P_msi.transpose((0, 2, 1)).copy().reshape((M, -1))
            for a, P_msi in self.projections.items()}

    def pdos_weights(self,
                     a: int,
                     indices: List[int]
                     ) -> Array3D:
        """PDOS weights."""
        dos_ms = np.zeros((self.projections.nbands, 2))

        P_amsi = self.projections

        if a in P_amsi:
            dos_ms[:, :] = (abs(P_amsi[a][:, :, indices])**2).sum(2)

        return dos_ms


class BZWaveFunctions:
    """Container for eigenvalues and PAW projections (all of BZ)."""
    def __init__(self,
                 kd: KPointDescriptor,
                 wfs: dict[int, WaveFunction],
                 occ: Optional[OccupationNumberCalculator],
                 nelectrons: float,
                 nl_aj: dict[int, list[tuple[int, int]]]):
        self.wfs = wfs
        self.occ = occ
        self.nelectrons = nelectrons
        self.nl_aj = nl_aj

        self.nbzkpts = kd.nbzkpts

        # Initialize ranks:
        self.ranks = np.zeros(kd.nbzkpts, int)
        for k in wfs:
            self.ranks[k] = kd.comm.rank
        kd.comm.sum(self.ranks)

        wf = next(iter(wfs.values()))  # get the first WaveFunction object

        self.shape = (kd.nbzkpts, wf.projections.nbands)
        self.domain_comm = wf.projections.atom_partition.comm
        self.bcomm = wf.projections.bcomm
        self.kpt_comm = kd.comm

        self.fermi_level = self._calculate_occ_numbers_and_fermi_level()

        self.size = kd.N_c
        self.bz2ibz_map = np.arange(self.nbzkpts)

    def weights(self):
        return np.zeros(len(self)) + 1 / self.nbzkpts

    def _calculate_occ_numbers_and_fermi_level(self) -> float:
        if self.occ is not None:
            eig_im = [wf.eig_m for wf in self]
            weight = 1.0 / self.nbzkpts
            weight_i = [weight] * len(eig_im)

            f_im, (fermi_level,), _ = self.occ.calculate(
                self.nelectrons,
                eig_im,
                weight_i)
            for wf, f_m in zip(self, f_im):
                wf.f_m[:] = f_m
        else:
            fermi_level = 0.0

        if self.domain_comm.rank == 0 and self.bcomm.rank == 0:
            fermi_level = self.bcomm.sum(fermi_level)
        fermi_level = self.domain_comm.sum(fermi_level)

        return fermi_level

    def calculate_band_energy(self) -> float:
        """Calculate sum over occupied eigenvalues."""
        if self.domain_comm.rank == 0 and self.bcomm.rank == 0:
            weight = 1.0 / self.nbzkpts
            e_band = sum(wf.eig_m.dot(wf.f_m) for wf in self) * weight
            e_band = self.kpt_comm.sum(e_band)
        else:
            e_band = 0.0

        if self.domain_comm.rank == 0:
            e_band = self.bcomm.sum(e_band)
        e_band = self.domain_comm.sum(e_band)
        return e_band

    def __iter__(self):
        for bz_index in sorted(self.wfs):
            yield self[bz_index]

    def __getitem__(self, bz_index):
        return self.wfs[bz_index]

    def __len__(self):
        return len(self.wfs)

    def eigenvalues(self,
                    broadcast: bool = True
                    ) -> Array2D:
        """Eigenvalues in eV for the whole BZ."""
        return self._collect(attrgetter('eig_m'), broadcast=broadcast)

    def occupation_numbers(self,
                           broadcast: bool = True
                           ) -> Array2D:
        """Occupation numbers for the whole BZ."""
        return self._collect(attrgetter('f_m'), broadcast=broadcast)

    def eigenvectors(self,
                     broadcast: bool = True
                     ) -> Array4D:
        """Eigenvectors for the whole BZ."""
        nbands = self.shape[1]
        assert nbands % 2 == 0
        return self._collect(attrgetter('v_mn'), (nbands,), complex,
                             broadcast=broadcast)

    def spin_projections(self,
                         broadcast: bool = True
                         ) -> Array3D:
        """Spin projections for the whole BZ."""
        return self._collect(attrgetter('spin_projection_mv'), (3,),
                             broadcast=broadcast)

    def get_orbital_magnetic_moments(self):
        """Return the orbital magnetic moment vector for each atom."""
        from gpaw.new.orbmag import get_orbmag_from_soc_eigs
        return get_orbmag_from_soc_eigs(self)

    def pdos_weights(self,
                     a: int,
                     indices: List[int],
                     broadcast: bool = True
                     ) -> Array4D:
        """Projections for PDOS.

        Returns (nbzkpts, nbands, 2)-shaped ndarray
        of the square of absolute value of the projections.
        """
        def func(wf):
            return wf.pdos_weights(a, indices)

        return self._collect(func,
                             (2,),
                             broadcast=broadcast,
                             sum_over_domain=True)

    def _collect(self,
                 func: Callable[[WaveFunction], ArrayND],
                 shape: Tuple[int, ...] = None,
                 dtype=float,
                 broadcast: bool = True,
                 sum_over_domain: bool = False) -> ArrayND:
        """Helper method for collecting (and broadcasting) ndarrays."""

        total_shape = self.shape + (shape or ())

        if broadcast:
            array_kmx = self._collect(func,
                                      shape,
                                      dtype,
                                      False,
                                      sum_over_domain)
            if array_kmx.ndim == 0:
                array_kmx = np.empty(total_shape, dtype)
            return broadcast_array(array_kmx,
                                   self.kpt_comm, self.bcomm, self.domain_comm)

        if self.bcomm.rank != 0:
            return np.empty(shape=())

        if not sum_over_domain and self.domain_comm.rank != 0:
            return np.empty(shape=())

        comm = self.kpt_comm
        if comm.rank == 0:
            array_kmx = np.empty(total_shape, dtype)
            for k, rank in enumerate(self.ranks):
                if rank == 0:
                    array_kmx[k] = func(self[k])
                else:
                    comm.receive(array_kmx[k], rank)
            if sum_over_domain:
                self.domain_comm.sum(array_kmx)
            if self.domain_comm.rank == 0:
                return array_kmx
            else:
                return np.empty(shape=())

        for k, rank in enumerate(self.ranks):
            if rank == comm.rank:
                comm.send(func(self[k]), 0)

        return np.empty(shape=())


def soc_eigenstates_raw(ibzwfs: Iterable[Tuple[int, WaveFunction]],
                        dVL_avii: Dict[int, Array3D],
                        ibz2bzmaps: IBZ2BZMaps,
                        atom_partition,
                        theta: float = 0.0,
                        phi: float = 0.0) -> Dict[int, WaveFunction]:

    theta *= np.pi / 180
    phi *= np.pi / 180

    # Hamiltonian with SO in KS basis
    # The even indices in H_mm are spin up along \hat n defined by \theta, phi
    # Basis change matrix for constructing Pauli matrices in \theta,\phi basis:
    #     \sigma_i^n = C^\dag\sigma_i C
    C_ss = np.array([[np.cos(theta / 2) * np.exp(-1.0j * phi / 2),
                      -np.sin(theta / 2) * np.exp(-1.0j * phi / 2)],
                     [np.sin(theta / 2) * np.exp(1.0j * phi / 2),
                      np.cos(theta / 2) * np.exp(1.0j * phi / 2)]])

    sx_ss = np.array([[0, 1], [1, 0]], complex)
    sy_ss = np.array([[0, -1.0j], [1.0j, 0]], complex)
    sz_ss = np.array([[1, 0], [0, -1]], complex)
    s_vss = [C_ss.T.conj() @ sx_ss @ C_ss,
             C_ss.T.conj() @ sy_ss @ C_ss,
             C_ss.T.conj() @ sz_ss @ C_ss]

    bzwfs = {}
    for ibz_index, ibzwf in ibzwfs:
        for K in np.nonzero(ibz2bzmaps.kd.bz2ibz_k == ibz_index)[0]:
            bzwf = ibzwf.transform(ibz2bzmaps[K], K)

            # Redistribute to match dVL_avii:
            bzwf = bzwf.redistribute_atoms(atom_partition)

            bzwf.add_soc(dVL_avii, s_vss, C_ss)
            bzwfs[K] = bzwf

    return bzwfs


def extract_ibz_wave_functions(kpt_qs: List[List[KPoint]],
                               bd: BandDescriptor,
                               gd: GridDescriptor,
                               n1: int,
                               n2: int,
                               eigenvalues: Array3D = None
                               ) -> Iterator[Tuple[int, WaveFunction]]:
    """Yield tuples of IBZ-index and wave functions.

    All atoms and bands will be on rank == 0 of gd.comm and bd.comm
    respectively.  This makes slicing the bands (from n1 to n2-1)
    and symmetry operations on the projections easier.
    """

    nproj_a = kpt_qs[0][0].projections.nproj_a

    collinear = kpt_qs[0][0].s is not None

    nbands = n2 - n1
    if collinear:
        nbands *= 2

    # All atoms on rank-0:
    atom_partition = AtomPartition(gd.comm, np.zeros_like(nproj_a))

    # All bands on rank-0 (nrow * ncol = 1 * 1):
    bdist = (bd.comm, 1, 1)

    for kpt_s in kpt_qs:
        # Collect bands and atoms to bd.comm.rank == 0 and gd.comm.rank == 0:
        if eigenvalues is None:
            eig_sn = [bd.collect(kpt.eps_n) for kpt in kpt_s]
        P_snI = [kpt.projections.collect() for kpt in kpt_s]

        projections = Projections(
            nbands=nbands,
            nproj_a=nproj_a,
            atom_partition=atom_partition,
            bdist=bdist,
            collinear=False)

        if bd.comm.rank == 0 and gd.comm.rank == 0:
            P1_nI = P_snI[0]
            P2_nI = P_snI[-1]
            assert P1_nI is not None
            assert P2_nI is not None
            if collinear:
                eig_m = np.empty((n2 - n1) * 2)
                if eigenvalues is None:
                    eig_m[::2] = eig_sn[0][n1:n2] * Ha
                    eig_m[1::2] = eig_sn[-1][n1:n2] * Ha
                else:
                    for s in range(2):
                        eig_m[s::2] = eigenvalues[-s, kpt_s[-s].k]
                projections.array[:] = 0.0
                projections.array[::2, 0] = P1_nI[n1:n2]
                projections.array[1::2, 1] = P2_nI[n1:n2]
            else:
                eig_m = eig_sn[0][n1:n2] * Ha
                projections.matrix.array[:] = P1_nI[n1:n2]
        else:
            eig_m = np.empty(0)

        ibz_index = kpt_s[0].k

        yield ibz_index, WaveFunction(eig_m, projections)


def soc_eigenstates(calc: ASECalculator | GPAW | str | Path,
                    n1: int = None,
                    n2: int = None,
                    scale: float = 1.0,
                    theta: float = 0.0,  # degrees
                    phi: float = 0.0,  # degrees
                    eigenvalues: Array3D = None,  # eV
                    occcalc: OccupationNumberCalculator = None,
                    projected: bool = False
                    ) -> BZWaveFunctions:
    """Calculate SOC eigenstates.

    Parameters:
        calc: Calculator
            GPAW calculator or path to gpw-file.
        n1, n2: int
            Range of bands to include (n1 <= n < n2).  Default is all
            bands available.
        scale: float
            Scale the spinorbit coupling by this amount.
        theta: float
            Angle in degrees.
        phi: float
            Angle in degrees.
        eigenvalues: ndarray
            Optionally use these eigenvalues instead for those from *calc*.
            The shape must be: (nspins, nibzkpts, n2 - n1).  Units: eV.
        occcalc:
            Occupation-number calculator.  By default, the one from *calc*
            will be used.

    Returns a BZWaveFunctions object covering the whole BZ.
    """

    from gpaw.calculator import GPAW  # noqa

    if isinstance(calc, (str, Path)):
        calc = GPAW(calc)

    n1 = n1 or 0
    n2 = n2 or 0
    if n2 <= 0:
        if eigenvalues is None:
            nbands = calc.get_number_of_bands()
        else:
            nbands = eigenvalues.shape[2]
        n2 += nbands

    # <phi_i|dV_adr / r * L_v|phi_j>
    dVL_avii = {a: soc(calc.wfs.setups[a],
                       calc.hamiltonian.xc, D_sp) * scale
                for a, D_sp in calc.density.D_asp.items()}

    if projected:
        dVL_avii = {a: projected_soc(dVL_vii, theta=theta, phi=phi)
                    for a, dVL_vii in dVL_avii.items()}

    kd = calc.wfs.kd
    bd = calc.wfs.bd
    gd = calc.wfs.gd
    atom_partition = calc.density.atom_partition

    if eigenvalues is not None:
        assert eigenvalues.shape == (kd.nspins, kd.nibzkpts, n2 - n1)

    ibzwfs = extract_ibz_wave_functions(calc.wfs.kpt_qs,
                                        bd, gd, n1, n2, eigenvalues)
    ibz2bzmaps = IBZ2BZMaps.from_calculator(calc)

    bzwfs = soc_eigenstates_raw(ibzwfs,
                                dVL_avii,
                                ibz2bzmaps,
                                atom_partition,
                                theta, phi)

    if bd.comm.rank == 0 and gd.comm.rank == 0:
        parallel_layout = ParallelLayout(BandDescriptor(1),
                                         kd.comm,
                                         serial_comm)
        occcalc = occcalc or calc.wfs.occupations
        occcalc = occcalc.copy(bz2ibzmap=list(range(kd.nbzkpts)),
                               parallel_layout=parallel_layout)
    else:
        occcalc = None

    nl_aj = {}
    for a, setup in enumerate(calc.wfs.setups):
        nl_aj[a] = list(zip(setup.n_j, setup.l_j))

    return BZWaveFunctions(kd, bzwfs, occcalc, calc.wfs.nvalence, nl_aj)


def soc(a: Setup, xc, D_sp: Array2D) -> Array3D:
    """<phi_i|dV_adr / r * L_v|phi_j>"""
    v_g = get_radial_potential(a, xc, D_sp)
    Ng = len(v_g)
    phi_jg = a.data.phi_jg

    Lx_lmm, Ly_lmm, Lz_lmm = get_L_vlmm()

    dVL_vii = np.zeros((3, a.ni, a.ni), complex)
    N1 = 0
    for j1, l1 in enumerate(a.l_j):
        Nm = 2 * l1 + 1
        N2 = 0
        for j2, l2 in enumerate(a.l_j):
            if l1 == l2:
                f_g = phi_jg[j1][:Ng] * v_g * phi_jg[j2][:Ng]
                c = a.xc_correction.rgd.integrate(f_g) / (4 * np.pi)
                dVL_vii[0, N1:N1 + Nm, N2:N2 + Nm] = c * Lx_lmm[l1]
                dVL_vii[1, N1:N1 + Nm, N2:N2 + Nm] = c * Ly_lmm[l1]
                dVL_vii[2, N1:N1 + Nm, N2:N2 + Nm] = c * Lz_lmm[l1]
            else:
                pass
            N2 += 2 * l2 + 1
        N1 += Nm
    return dVL_vii * alpha**2 / 4.0


def projected_soc(dVL_vii: Array3D,
                  theta: float = 0,
                  phi: float = 0) -> Array3D:
    """
    Optional Args:
        theta (float): The angle from z-axis in degrees
        phi (float): The angle from x-axis in degrees
    """
    theta *= np.pi / 180
    phi *= np.pi / 180
    n_v = np.array([np.sin(theta) * np.cos(phi),
                    np.sin(theta) * np.sin(phi),
                    np.cos(theta)])
    dVL_vii = (np.dot(dVL_vii.T, n_v)[:, :, np.newaxis] * n_v).T
    return dVL_vii


def get_radial_potential(a: Setup, xc, D_sp: Array2D) -> Array1D:
    """Calculates (dV/dr)/r for the effective potential.
    Below, f_g denotes dV/dr = minus the radial force"""

    rgd = a.xc_correction.rgd
    r_g = rgd.r_g.copy()
    r_g[0] = 1.0e-12
    dr_g = rgd.dr_g

    B_pq = a.xc_correction.B_pqL[:, :, 0]
    n_qg = a.xc_correction.n_qg
    D_sq = np.dot(D_sp, B_pq)
    n_sg = np.dot(D_sq, n_qg) / (4 * np.pi)**0.5
    Ns = len(D_sp)
    if Ns == 4:
        Ns = 1
    n_sg[:Ns] += a.xc_correction.nc_g / Ns

    # Coulomb force from nucleus
    fc_g = a.Z / r_g**2

    # Hartree force
    rho_g = 4 * np.pi * r_g**2 * dr_g * np.sum(n_sg[:Ns], axis=0)
    fh_g = -np.array([np.sum(rho_g[:ig]) for ig in range(len(r_g))]) / r_g**2

    f_g = fc_g + fh_g

    # xc force
    if xc.type != 'GLLB':
        v_sg = np.zeros_like(n_sg)
        xc.calculate_spherical(rgd, n_sg, v_sg)
        fxc_g = np.mean([rgd.derivative(v_g) for v_g in v_sg[:Ns]],
                        axis=0)
        f_g += fxc_g

    return f_g / r_g


def get_anisotropy(calc, theta=0.0, phi=0.0, nbands=0, width=None):
    """Calculates the sum of occupied spinorbit eigenvalues.

    Returns the result relative to the sum of eigenvalues without
    spinorbit coupling.
    """
    raise RuntimeError('Please use BZWaveFunctions.calculate_band_energy() '
                       'instead.')


def get_magnetic_moments(calc, theta=0.0, phi=0.0, nbands=None, width=None):
    """Calculates the magnetic moments inside all PAW spheres"""

    raise RuntimeError(
        'This function has no tests.  It is very likely that it no longer '
        'works correctly after merging !677.')

    from gpaw.utilities import unpack

    if nbands is None:
        nbands = calc.get_number_of_bands()
    Nk = len(calc.get_ibz_k_points())

    C_ss = np.array([[np.cos(theta / 2) * np.exp(-1.0j * phi / 2),
                      -np.sin(theta / 2) * np.exp(-1.0j * phi / 2)],
                     [np.sin(theta / 2) * np.exp(1.0j * phi / 2),
                      np.cos(theta / 2) * np.exp(1.0j * phi / 2)]])
    sx_ss = np.array([[0, 1], [1, 0]], complex)
    sy_ss = np.array([[0, -1.0j], [1.0j, 0]], complex)
    sz_ss = np.array([[1, 0], [0, -1]], complex)
    sx_ss = C_ss.T.conj().dot(sx_ss).dot(C_ss)
    sy_ss = C_ss.T.conj().dot(sy_ss).dot(C_ss)
    sz_ss = C_ss.T.conj().dot(sz_ss).dot(C_ss)

    states = soc_eigenstates(calc,
                             theta=theta,
                             phi=phi,
                             return_wfs=True,
                             bands=range(nbands))
    e_km = states['eigenvalues']
    v_knm = states['eigenstates']

    from gpaw.occupations import occupation_numbers
    if width is None:
        assert calc.wfs.occupations.name == 'fermi-dirac'
        width = calc.wfs.occupations._width
    if width == 0.0:
        width = 1.e-6
    weight_k = calc.get_k_point_weights() / 2
    ne = calc.wfs.setups.nvalence - calc.density.charge
    f_km = occupation_numbers({'name': 'fermi-dirac', 'width': width},
                              e_km[np.newaxis],
                              weight_k=weight_k,
                              nelectrons=ne)[0][0]

    m_v = np.zeros(3, complex)
    for ik in range(Nk):
        ut0_nG = np.array([calc.wfs.get_wave_function_array(n, ik, 0)
                           for n in range(nbands)])
        ut1_nG = np.array([calc.wfs.get_wave_function_array(n, ik, 1)
                           for n in range(nbands)])
        mocc = np.where(f_km[ik] * Nk - 1.0e-6 < 0.0)[0][0]
        for m in range(mocc + 1):
            f = f_km[ik, m]
            ut0_G = np.dot(v_knm[ik][::2, m], np.swapaxes(ut0_nG, 0, 2))
            ut1_G = np.dot(v_knm[ik][1::2, m], np.swapaxes(ut1_nG, 0, 2))
            ut_sG = np.array([ut0_G, ut1_G])

            mx_G = np.zeros(np.shape(ut0_G), complex)
            my_G = np.zeros(np.shape(ut0_G), complex)
            mz_G = np.zeros(np.shape(ut0_G), complex)
            for s1 in range(2):
                for s2 in range(2):
                    mx_G += ut_sG[s1].conj() * sx_ss[s1, s2] * ut_sG[s2]
                    my_G += ut_sG[s1].conj() * sy_ss[s1, s2] * ut_sG[s2]
                    mz_G += ut_sG[s1].conj() * sz_ss[s1, s2] * ut_sG[s2]
            m_v += calc.wfs.gd.integrate(np.array([mx_G, my_G, mz_G])) * f

    m_av = []
    for a in range(len(calc.atoms)):
        N0_p = calc.density.setups[a].N0_p.copy()
        N0_ij = unpack(N0_p)
        Dx_ij = np.zeros_like(N0_ij, complex)
        Dy_ij = np.zeros_like(N0_ij, complex)
        Dz_ij = np.zeros_like(N0_ij, complex)
        Delta_p = calc.density.setups[a].Delta_pL[:, 0].copy()
        Delta_ij = unpack(Delta_p)
        for ik in range(Nk):
            P_ami = ...  # get_spinorbit_projections(calc, ik, v_knm[ik])
            P_smi = np.array([P_ami[a][:, ::2], P_ami[a][:, 1::2]])
            P_smi = np.dot(C_ss, np.swapaxes(P_smi, 0, 1))

            P0_mi = P_smi[0]
            P1_mi = P_smi[1]
            f_mm = np.diag(f_km[ik])

            Dx_ij += P0_mi.conj().T.dot(f_mm).dot(P1_mi)
            Dx_ij += P1_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dy_ij -= 1.0j * P0_mi.conj().T.dot(f_mm).dot(P1_mi)
            Dy_ij += 1.0j * P1_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dz_ij += P0_mi.conj().T.dot(f_mm).dot(P0_mi)
            Dz_ij -= P1_mi.conj().T.dot(f_mm).dot(P1_mi)

        mx = np.sum(N0_ij * Dx_ij).real
        my = np.sum(N0_ij * Dy_ij).real
        mz = np.sum(N0_ij * Dz_ij).real

        m_av.append([mx, my, mz])
        m_v[0] += np.sum(Delta_ij * Dx_ij) * (4 * np.pi)**0.5
        m_v[1] += np.sum(Delta_ij * Dy_ij) * (4 * np.pi)**0.5
        m_v[2] += np.sum(Delta_ij * Dz_ij) * (4 * np.pi)**0.5

    return m_v.real, m_av


def get_parity_eigenvalues(calc, ik=0, spin_orbit=False, bands=None, Nv=None,
                           inversion_center=[0, 0, 0], deg_tol=1.0e-6,
                           tol=1.0e-6):
    """Calculates parity eigenvalues at time-reversal invariant k-points.
    Only works in plane wave mode.
    """

    assert len(calc.get_bz_k_points()) == len(calc.get_ibz_k_points())

    kpt_c = calc.get_ibz_k_points()[ik]
    if Nv is None:
        Nv = int(calc.get_number_of_electrons() / 2)

    if bands is None:
        bands = range(calc.get_number_of_bands())

    # Find degenerate subspaces
    eig_n = calc.get_eigenvalues(kpt=ik)[bands]
    e_in = []
    used_n = []
    for n1, e1 in enumerate(eig_n):
        if n1 not in used_n:
            n_n = []
            for n2, e2 in enumerate(eig_n):
                if np.abs(e1 - e2) < deg_tol:
                    n_n.append(n2)
                    used_n.append(n2)
            e_in.append(n_n)

    print()
    print(f' Inversion center at: {inversion_center}')
    print(f' Calculating inversion eigenvalues at k = {kpt_c}')
    print()

    center_v = np.array(inversion_center) / Bohr
    G_Gv = calc.wfs.pd.get_reciprocal_vectors(q=ik, add_q=True)

    psit_nG = np.array([calc.wfs.kpt_u[ik].psit_nG[n]
                        for n in bands])
    if spin_orbit:
        n1 = bands[0]
        n2 = bands[-1] + 1
        assert (bands == np.arange(n1, n2)).all()
        soc = soc_eigenstates(calc, n1=n1, n2=n2)
        v_kmn = soc.eigenvectors()
        psit0_mG = np.dot(v_kmn[ik, :, ::2], psit_nG)
        psit1_mG = np.dot(v_kmn[ik, :, 1::2], psit_nG)
    for n in range(len(bands)):
        psit_nG[n] /= (np.sum(np.abs(psit_nG[n])**2))**0.5
    if spin_orbit:
        for n in range(2 * len(bands)):
            A = np.sum(np.abs(psit0_mG[n])**2) + np.sum(np.abs(psit1_mG[n])**2)
            psit0_mG[n] /= A**0.5
            psit1_mG[n] /= A**0.5

    P_GG = np.ones((len(G_Gv), len(G_Gv)), float)
    for iG, G_v in enumerate(G_Gv):
        P_GG[iG] -= ((G_Gv[:] + G_v).round(6)).any(axis=1)
    assert (P_GG == P_GG.T).all()

    phase_G = np.exp(-2.0j * np.dot(G_Gv, center_v))

    p_n = []
    print('n   P_n')
    for n_n in e_in:
        if spin_orbit:
            # The dimension of parity matrix is doubled with spinorbit
            m_m = [2 * n_n[0] + i for i in range(2 * len(n_n))]
            Ppsit0_mG = np.dot(P_GG, psit0_mG[m_m].T).T
            Ppsit0_mG[:] *= phase_G
            Ppsit1_mG = np.dot(P_GG, psit1_mG[m_m].T).T
            Ppsit1_mG[:] *= phase_G
            P_nn = np.dot(psit0_mG[m_m].conj(), np.array(Ppsit0_mG).T)
            P_nn += np.dot(psit1_mG[m_m].conj(), np.array(Ppsit1_mG).T)
        else:
            Ppsit_nG = np.dot(P_GG, psit_nG[n_n].T).T
            Ppsit_nG[:] *= phase_G
            P_nn = np.dot(psit_nG[n_n].conj(), np.array(Ppsit_nG).T)
        P_eig = np.linalg.eigh(P_nn)[0]
        if np.allclose(np.abs(P_eig), 1, tol):
            P_n = np.sign(P_eig).astype(int).tolist()
            if spin_orbit:
                # Only include one of the degenerate pair of eigenvalues
                Pm = np.sign(P_eig).tolist().count(-1)
                Pp = np.sign(P_eig).tolist().count(1)
                P_n = Pm // 2 * [-1] + Pp // 2 * [1]
            print(f'{str(n_n)[1:-1]}: {str(P_n)[1:-1]}')
            p_n += P_n
        else:
            print(f'  {n_n} are not parity eigenstates')
            print(f'     P_n: {P_eig}')
            print(f'     e_n: {eig_n[n_n]}')
            p_n += [0 for n in n_n]

    return np.ravel(p_n)


def get_L_vlmm():
    if len(_L_vlmm) == 3:
        return _L_vlmm

    s = np.array([[0.0]])
    p = np.zeros((3, 3), complex)  # y, z, x
    p[0, 1] = -1.0j
    p[1, 0] = 1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 3] = -1.0j
    d[3, 0] = 1.0j
    d[1, 2] = -3**0.5 * 1.0j
    d[2, 1] = 3**0.5 * 1.0j
    d[1, 4] = -1.0j
    d[4, 1] = 1.0j
    _L_vlmm.append([s, p, d])

    p = np.zeros((3, 3), complex)  # y, z, x
    p[1, 2] = -1.0j
    p[2, 1] = 1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 1] = 1.0j
    d[1, 0] = -1.0j
    d[2, 3] = -3**0.5 * 1.0j
    d[3, 2] = 3**0.5 * 1.0j
    d[3, 4] = -1.0j
    d[4, 3] = 1.0j
    _L_vlmm.append([s, p, d])

    p = np.zeros((3, 3), complex)  # y, z, x
    p[0, 2] = 1.0j
    p[2, 0] = -1.0j
    d = np.zeros((5, 5), complex)  # xy, yz, z^2, xz, x^2-y^2
    d[0, 4] = 2.0j
    d[4, 0] = -2.0j
    d[1, 3] = 1.0j
    d[3, 1] = -1.0j
    _L_vlmm.append([s, p, d])

    return _L_vlmm



from packaging.version import Version
from gpaw import __ase_version_required__
from ase import __version__


def test_ase_features_ase3k_version():
    assert Version(__version__) >= Version(__ase_version_required__)



"""This test looks for differently named setups and basis sets to ascertain
that correct filenames are constructed.  Generally speaking, the things
it looks for do not exist; we just verify the filenames.
"""
import pytest
from gpaw.basis_data import get_basis_name, parse_basis_name
from gpaw.setup import Setups


@pytest.mark.ci
@pytest.mark.parametrize(
    'setups, basis, refname',
    [({}, 'hello', 'Na.hello.basis'),
     ('hello', {}, 'Na.hello.LDA'),
     ('hello', 'dzp', 'Na.hello.dzp.basis'),
     ('hello', 'sz(dzp)', 'Na.hello.dzp.basis'),
     ('hello', 'world.dzp', 'Na.hello.world.dzp.basis'),
     ('hello', 'sz(world.dzp)', 'Na.hello.world.dzp.basis'),
     ('paw', 'world.dzp', 'Na.world.dzp.basis'),
     ('paw', 'sz(world.dzp)', 'Na.world.dzp.basis')])
def test_setup_basis_spec(setups, basis, refname):
    with pytest.raises(FileNotFoundError) as err:
        Setups([11], setups, basis, 'LDA')
    msg = str(err)
    fname = msg.split('"')[1]
    assert fname == refname, fname


@pytest.mark.parametrize('basisname',
                         'sz dz szp dzp dz2p dzdp tzqp dzsp'.split())
def test_basis_something_something(basisname):
    zetacount, polarizationcount = parse_basis_name(basisname)
    normalized_name = get_basis_name(zetacount, polarizationcount)
    zetacount2, polarizationcount2 = parse_basis_name(normalized_name)
    assert zetacount == zetacount2
    assert polarizationcount == polarizationcount2
    if (polarizationcount < 2 and
        len(basisname) < 4 and basisname.isalpha()):
        assert normalized_name == basisname



"""Check that the LCAODOS works."""

import pytest
from ase.build import molecule

from gpaw import GPAW


@pytest.mark.later
def test_lcao_dos():
    system = molecule('H2O')
    system.center(vacuum=3.0)
    system.pbc = 1

    calc = GPAW(mode='lcao', basis='dzp', h=0.3, xc='oldLDA',
                kpts=[2, 1, 1],  # parallel=dict(sl_auto=True),
                nbands=8,
                # kpts=[4, 4, 4]
                )

    system.calc = calc
    system.get_potential_energy()

    from gpaw.utilities.dos import LCAODOS

    # Use RestartLCAODOS if you just restarted from a file.
    # Requires one diagonalization though!

    lcaodos = LCAODOS(calc)

    def printdos(eps, w):
        print('sum of weights', sum(w))
        print('energy, weight')
        for e0, w0 in zip(eps, w):
            print(e0, w0)
        print('-----')
    eps0, w0 = lcaodos.get_orbital_pdos(0)  # O 2s

    printdos(eps0, w0)

    assert w0[0] > 1.6  # s state on which we projected with two electrons.
    assert (w0[1:] < 0.2).all()  # there is another s with some weight ~ 0.16

    # eps1_n, w1_n = fold(eps * 27.211, w, 600, 0.1)
    # pl.plot(eps1, w1)
    # pl.show()

    # for e0, w0 in zip(eps, w):
    #     print e0, w0

    # print eps.shape
    # print w.shape

    a1 = [0, 1, 2]
    eps1, w1 = lcaodos.get_atomic_subspace_pdos(a1)
    assert abs(2 * len(eps1) - sum(w1)) < 0.01

    printdos(eps1, w1)
    print('indices %s' % lcaodos.get_atom_indices(a1))

    a2 = 0  # all BFs on the O atom
    eps2, w2 = lcaodos.get_atomic_subspace_pdos(a2)
    printdos(eps2, w2)
    print('indices %s' % lcaodos.get_atom_indices(a2))



import pytest
from ase import Atoms
from ase.units import Ha

from gpaw.new.calculation import DFTCalculation
from gpaw.mpi import size


@pytest.mark.gpu
@pytest.mark.serial
@pytest.mark.parametrize('dtype', [float, complex])
@pytest.mark.parametrize('gpu', [False, True])
def test_gpu_pw(dtype, gpu):
    if gpu and dtype == float:
        pytest.skip('P_ani * dH_aii kernel not implemented for float')
    atoms = Atoms('H2')
    atoms.positions[1, 0] = 0.75
    atoms.center(vacuum=1.0)
    dft = DFTCalculation.from_parameters(
        atoms,
        dict(mode={'name': 'pw'},
             dtype=dtype,
             parallel={'gpu': gpu},
             setups='paw'),
        log='-')
    dft.converge()
    dft.energies()
    energy = dft.results['energy'] * Ha
    assert energy == pytest.approx(-16.032945, abs=1e-6)


@pytest.mark.gpu
@pytest.mark.skipif(size > 2, reason='Not implemented')
@pytest.mark.parametrize('gpu', [False, True])
@pytest.mark.parametrize('par', ['domain', 'kpt'])
def test_gpu_pw_k(gpu, par):
    atoms = Atoms('H', pbc=True, cell=[1, 1, 1])
    dft = DFTCalculation.from_parameters(
        atoms,
        dict(mode={'name': 'pw'},
             kpts=(4, 1, 1),
             parallel={'gpu': gpu,
                       par: size},
             setups='paw'),
        log='-')
    dft.converge()
    dft.energies()
    dft.forces()
    if par == 'kpt':
        dft.stress()
    energy = dft.results['energy'] * Ha
    assert energy == pytest.approx(-19.579937, abs=1e-6)



import numpy as np
from ase.utils.timing import timer

from gpaw.lcao.eigensolver import DirectLCAO
from gpaw.lfc import BasisFunctions
from gpaw.matrix import Matrix, matrix_matrix_multiply as mmm
from gpaw.utilities import unpack
from gpaw.utilities.timing import nulltimer
from gpaw.wavefunctions.base import WaveFunctions
from gpaw.wavefunctions.lcao import LCAOWaveFunctions, update_phases


class NullWfsMover:
    description = 'Wavefunctions kept unchanged if atoms move'

    def initialize(self, lcaowfs):
        pass

    def cut_wfs(self, wfs, spos_ac):
        pass


class PseudoPartialWaveWfsMover:
    """Move wavefunctions with atoms according to PAW basis

    Wavefunctions are approximated around atom a as

       ~          --  ~ a      ~a    ~
      psi (r)  ~  >  phi (r) < p  | psi >
         n        --    i       i      n
                  ai

    This quantity is then subtracted and re-added at the new
    positions.
    """
    description = 'Improved wavefunction reuse through dual PAW basis'

    def initialize(self, lcaowfs):
        pass

    def cut_wfs(self, wfs, spos_ac):
        ni_a = {}

        for a in range(len(wfs.setups)):
            setup = wfs.setups[a]
            l_j = [phit.get_angular_momentum_number()
                   for phit in setup.get_partial_waves_for_atomic_orbitals()]
            # assert l_j == setup.l_j[:len(l_j)]  # Relationship to l_orb_J?
            ni_a[a] = sum(2 * l + 1 for l in l_j)

        phit = wfs.get_pseudo_partial_waves()
        phit.set_positions(wfs.spos_ac, wfs.atom_partition)

        # XXX See also wavefunctions.lcao.update_phases
        if wfs.dtype == complex:
            phase_qa = np.exp(2j * np.pi *
                              np.dot(wfs.kd.ibzk_qc,
                                     (spos_ac - wfs.spos_ac).T.round()))

        def add_phit_to_wfs(multiplier):
            for kpt in wfs.kpt_u:
                P_ani = {}
                for a in kpt.P_ani:
                    P_ani[a] = multiplier * kpt.P_ani[a][:, :ni_a[a]]
                    if multiplier > 0 and wfs.dtype == complex:
                        P_ani[a] *= phase_qa[kpt.q, a]
                phit.add(kpt.psit_nG, c_axi=P_ani, q=kpt.q)

        add_phit_to_wfs(-1.0)

        def paste():
            phit.set_positions(spos_ac, wfs.atom_partition)
            add_phit_to_wfs(1.0)

        return paste


class LCAOWfsMover:
    """Move wavefunctions with atoms according to LCAO basis.

    Approximate wavefunctions as a linear combination of atomic
    orbitals, then subtract that linear combination and re-add
    it after moving the atoms using the same coefficients.

    The coefficients c are determined by the equation

               /    *  _  ^  ~   _   _    --
      X     =  | Phi  (r) S psi (r) dr =  >  S      c
       n mu    /    mu         n          --  mu nu  nu n
                                          nu

    We calculate X directly and then solve for c.
    """
    description = 'Improved wavefunction reuse through full LCAO basis'

    # TODO/FIXME
    # * Get rid of the unnecessary T matrix
    # * Only recalculate S/P when necessary (not first time)
    # * Full parallelization support (ScaLAPACK, check scipy atomic correction)
    #   Also replace np.linalg.solve by parallel/efficient thing
    # * Broken with PW mode because PW mode has very funny P_ani shapes.
    #   Also PW does not use the overlap object; this may be related
    # * Can we use updated S matrix to construct better guess?

    def initialize(self, lcaowfs):
        self.bfs = lcaowfs.basis_functions
        # self.tci = lcaowfs.tci
        self.tciexpansions = lcaowfs.tciexpansions
        self.atomic_correction = lcaowfs.atomic_correction
        # self.S_qMM = lcaowfs.S_qMM
        # self.T_qMM = lcaowfs.T_qMM  # Get rid of this
        # self.P_aqMi = lcaowfs.P_aqMi

    def cut_wfs(self, wfs, spos_ac):
        # XXX Must forward vars from LCAO initialization object
        # in order to not need to recalculate them.
        # Also, if we get the vars from the LCAO init object,
        # we can rely on those parallelization settings without danger.
        bfs = self.bfs

        # P_aqMi = self.P_aqMi
        # S_qMM = self.S_qMM

        # We can inherit S_qMM and P_aqMi from the initialization in the
        # first step, then recalculate them for subsequent steps.
        wfs.timer.start('reuse wfs')
        wfs.timer.start('tci calculate')
        tciex = self.tciexpansions
        manytci = tciex.get_manytci_calculator(wfs.setups, wfs.gd,
                                               spos_ac, wfs.kd.ibzk_qc,
                                               wfs.dtype, wfs.timer)
        P_aqMi = manytci.P_aqMi(bfs.my_atom_indices)
        # Avoid calculating T
        Mstart, Mstop = wfs.initksl.Mstart, wfs.initksl.Mstop
        S_qMM, T_qMM = manytci.O_qMM_T_qMM(wfs.gd.comm, Mstart, Mstop)

        oldcorr = self.atomic_correction
        from gpaw.lcao.atomic_correction import DenseAtomicCorrection
        corr = DenseAtomicCorrection(P_aqMi, oldcorr.dS_aii, Mstart, Mstop)

        wfs.timer.stop('tci calculate')
        # self.atomic_correction.initialize(P_aqMi, Mstart, Mstop)
        # self.atomic_correction.gobble_data(wfs)
        wfs.timer.start('lcao overlap correction')
        corr.add_overlap_correction(S_qMM)
        # self.atomic_correction.add_overlap_correction(S_qMM)
        wfs.timer.stop('lcao overlap correction')
        wfs.gd.comm.sum(S_qMM)
        c_unM = []
        for kpt in wfs.kpt_u:
            S_MM = S_qMM[kpt.q]
            X_nM = np.zeros((wfs.bd.mynbands, wfs.setups.nao), wfs.dtype)
            # XXX use some blocksize to reduce memory usage?
            with wfs.timer('wfs overlap'):
                opsit = kpt.psit.new()
                opsit.array[:] = kpt.psit.array
                opsit.add(wfs.pt, wfs.setups.dS.apply(kpt.projections))

            with wfs.timer('bfs integrate'):
                bfs.integrate2(opsit.array, c_xM=X_nM, q=kpt.q)

            with wfs.timer('gd comm sum'):
                wfs.gd.comm.sum(X_nM)

            # Mind band parallelization / ScaLAPACK
            # Actually we can probably ignore ScaLAPACK for FD/PW calculations
            # since we never adapted Davidson to those.  Although people
            # may have requested ScaLAPACK for LCAO initialization.
            c_nM = np.linalg.solve(S_MM.T, X_nM.T).T.copy()

            # c_nM *= 0  # This disables the whole mechanism
            with wfs.timer('lcao to grid'):
                bfs.lcao_to_grid(C_xM=-c_nM, psit_xG=kpt.psit_nG, q=kpt.q)

            c_unM.append(c_nM)

        if wfs.dtype == complex:
            update_phases(c_unM, [kpt.q for kpt in wfs.kpt_u], wfs.kd.ibzk_qc,
                          spos_ac, wfs.spos_ac, wfs.setups, wfs.initksl.Mstart)

        del opsit

        with wfs.timer('bfs set pos'):
            bfs.set_positions(spos_ac)

        # Is it possible to recalculate the overlaps and make use of how
        # they have changed here?
        wfs.timer.start('re-add wfs')
        for u, kpt in enumerate(wfs.kpt_u):
            bfs.lcao_to_grid(C_xM=c_unM[u], psit_xG=kpt.psit_nG, q=kpt.q)
        wfs.timer.stop('re-add wfs')
        wfs.timer.stop('reuse wfs')


class FDPWWaveFunctions(WaveFunctions):
    """Base class for finite-difference and planewave classes."""
    def __init__(self, parallel, initksl, reuse_wfs_method=None, **kwargs):
        WaveFunctions.__init__(self, **kwargs)

        ranks = [rbd * self.gd.comm.size + rgd
                 for rgd in range(self.gd.comm.size)
                 for rbd in range(self.bd.comm.size)]
        slcomm = parallel[0].new_communicator(ranks)
        self.scalapack_parameters = (slcomm, ) + parallel[1:]

        self.initksl = initksl
        if reuse_wfs_method is None or reuse_wfs_method == 'keep':
            wfs_mover = NullWfsMover()
        elif hasattr(reuse_wfs_method, 'cut_wfs'):
            wfs_mover = reuse_wfs_method
        elif reuse_wfs_method == 'paw':
            wfs_mover = PseudoPartialWaveWfsMover()
        elif reuse_wfs_method == 'lcao':
            wfs_mover = LCAOWfsMover()
        else:
            raise ValueError('Strange way to reuse wfs: {}'
                             .format(reuse_wfs_method))

        self.wfs_mover = wfs_mover

        self.set_orthonormalized(False)

        self._work_matrix_nn = None  # storage for H, S, ...
        self._work_array = None

    @property
    def work_array(self):
        if self._work_array is None:
            self._work_array = self.empty(self.bd.mynbands *
                                          (1 if self.collinear else 2))
        return self._work_array

    @property
    def work_matrix_nn(self):
        """Get Matrix object for H, S, ..."""
        if self._work_matrix_nn is None:
            self._work_matrix_nn = Matrix(
                self.bd.nbands, self.bd.nbands,
                self.dtype,
                dist=(self.bd.comm, self.bd.comm.size))
        return self._work_matrix_nn

    def __str__(self):
        comm, r, c, b = self.scalapack_parameters
        L1 = ('  ScaLapack parameters: grid={}x{}, blocksize={}'
              .format(r, c, b))
        L2 = ('  Wavefunction extrapolation:\n    {}'
              .format(self.wfs_mover.description))
        return '\n'.join([L1, L2])

    def set_setups(self, setups):
        WaveFunctions.set_setups(self, setups)

    def set_orthonormalized(self, flag):
        self.orthonormalized = flag

    def set_positions(self, spos_ac, atom_partition=None):
        # We don't move the wave functions if psit_nG is None or an
        # NDArrayReader instance (and we need positions also):
        move_wfs = (isinstance(self.kpt_u[0].psit_nG, np.ndarray) and
                    self.spos_ac is not None)

        if move_wfs:
            paste_wfs = self.wfs_mover.cut_wfs(self, spos_ac)

        # This will update the positions -- and transfer, if necessary --
        # the projection matrices which may be necessary for updating
        # the wavefunctions.
        WaveFunctions.set_positions(self, spos_ac, atom_partition)

        if move_wfs and paste_wfs is not None:
            paste_wfs()

        self.set_orthonormalized(False)
        self.pt.set_positions(spos_ac, atom_partition)
        self.allocate_arrays_for_projections(self.pt.my_atom_indices)
        self.positions_set = True

    def initialize(self, density, hamiltonian, spos_ac):
        """Initialize wave-functions, density and hamiltonian.

        Return (nlcao, nrand) tuple with number of bands intialized from
        LCAO and random numbers, respectively."""

        if self.kpt_u[0].psit is None:
            basis_functions = BasisFunctions(self.gd,
                                             [setup.basis_functions_J
                                              for setup in self.setups],
                                             self.kd, dtype=self.dtype,
                                             cut=True)
            basis_functions.set_positions(spos_ac)
        else:
            self.initialize_wave_functions_from_restart_file()

        if self.kpt_u[0].psit is not None:
            density.initialize_from_wavefunctions(self)
        elif density.nt_sG is None:
            density.initialize_from_atomic_densities(basis_functions)
            # Initialize GLLB-potential from basis function orbitals
            if hamiltonian.xc.type == 'GLLB':
                hamiltonian.xc.initialize_from_atomic_orbitals(
                    basis_functions)
        else:  # XXX???
            # We didn't even touch density, but some combinations in paw.set()
            # will make it necessary to do this for some reason.
            density.calculate_normalized_charges_and_mix()
        hamiltonian.update(density)

        if self.kpt_u[0].psit is None:
            if 1:  # self.collinear:
                nlcao = self.initialize_wave_functions_from_basis_functions(
                    basis_functions, density, hamiltonian, spos_ac)
            else:
                self.random_wave_functions(0)
                nlcao = 0
            nrand = self.bd.nbands - nlcao
        else:
            # We got everything from file:
            nlcao = 0
            nrand = 0

        return nlcao, nrand

    def initialize_wave_functions_from_restart_file(self):
        for kpt in self.kpt_u:
            if not kpt.psit.in_memory:
                kpt.psit.read_from_file()

    def initialize_wave_functions_from_basis_functions(self,
                                                       basis_functions,
                                                       density, hamiltonian,
                                                       spos_ac):
        # if self.initksl is None:
        #     raise RuntimeError('use fewer bands or more basis functions')

        self.timer.start('LCAO initialization')
        lcaoksl, lcaobd = self.initksl, self.initksl.bd
        lcaowfs = LCAOWaveFunctions(lcaoksl, self.gd, self.nvalence,
                                    self.setups, lcaobd, self.dtype,
                                    self.world, self.kd, self.kptband_comm,
                                    nulltimer)
        lcaowfs.basis_functions = basis_functions
        lcaowfs.timer = self.timer
        self.timer.start('Set positions (LCAO WFS)')
        lcaowfs.set_positions(spos_ac, self.atom_partition)
        self.timer.stop('Set positions (LCAO WFS)')

        if self.collinear:
            eigensolver = DirectLCAO()
        else:
            from gpaw.xc.noncollinear import NonCollinearLCAOEigensolver
            eigensolver = NonCollinearLCAOEigensolver()

        eigensolver.initialize(self.gd, self.dtype, self.setups.nao, lcaoksl)

        # XXX when density matrix is properly distributed, be sure to
        # update the density here also
        eigensolver.iterate(hamiltonian, lcaowfs)

        # Transfer coefficients ...
        for kpt, lcaokpt in zip(self.kpt_u, lcaowfs.kpt_u):
            kpt.C_nM = lcaokpt.C_nM

        self.wfs_mover.initialize(lcaowfs)

        # and get rid of potentially big arrays early:
        del eigensolver, lcaowfs

        with self.timer('LCAO to grid'):
            self.initialize_from_lcao_coefficients(basis_functions)

        if not self.collinear and self.bd.mynbands > lcaobd.mynbands * 2:
            raise ValueError('Too many bands for non-collinear calculation!')

        if self.collinear and self.bd.mynbands > lcaobd.mynbands:
            # Add extra states.  If the number of atomic orbitals is
            # less than the desired number of bands, then extra random
            # wave functions are added.
            self.random_wave_functions(lcaobd.mynbands)
            # IMPORTANT: This intersperses random wavefunctions
            # with those from LCAO depending on band parallelization.
            # This is presumably okay as long as the FD/PW eigensolver
            # is called again before using the wavefunctions/occupations.
            #
            # Indeed as of writing this, the initialization appears to
            # call these things in the correct order, but there is no
            # telling when this will break due to some unrelated change.
        self.timer.stop('LCAO initialization')

        return lcaobd.nbands

    @timer('Orthonormalize')
    def orthonormalize(self, kpt=None):
        if kpt is None:
            for kpt in self.kpt_u:
                self.orthonormalize(kpt)
            self.orthonormalized = True
            return

        psit = kpt.psit
        P = kpt.projections

        with self.timer('projections'):
            psit.matrix_elements(self.pt, out=P)

        S = self.work_matrix_nn
        P2 = P.new()

        with self.timer('calc_s_matrix'):
            psit.matrix_elements(out=S, symmetric=True, cc=True)
            self.setups.dS.apply(P, out=P2)
            mmm(1.0, P, 'N', P2, 'C', 1.0, S, symmetric=True)

        with self.timer('inverse-cholesky'):
            S.invcholesky()
            # S now contains the inverse of the Cholesky factorization

        psit2 = psit.new(buf=self.work_array)
        with self.timer('rotate_psi_s'):
            mmm(1.0, S, 'N', psit, 'N', 0.0, psit2)
            mmm(1.0, S, 'N', P, 'N', 0.0, P2)
            psit[:] = psit2
            kpt.projections = P2

    def calculate_forces(self, hamiltonian, F_av):
        # Calculate force-contribution from k-points:
        F_av[:] = 0.0

        if not self.collinear:
            F_ansiv = self.pt.dict(2 * self.bd.mynbands, derivative=True)
            dH_axp = hamiltonian.dH_asp
            for kpt in self.kpt_u:
                array = kpt.psit.array
                self.pt.derivative(array.reshape((-1, array.shape[-1])),
                                   F_ansiv, kpt.q)
                for a, F_nsiv in F_ansiv.items():
                    F_nsiv = F_nsiv.reshape((self.bd.mynbands,
                                             2, -1, 3)).conj()
                    F_nsiv *= kpt.f_n[:, np.newaxis, np.newaxis, np.newaxis]
                    dH_ii = unpack(dH_axp[a][0])
                    dH_vii = [unpack(dH_p) for dH_p in dH_axp[a][1:]]
                    dH_ssii = np.array(
                        [[dH_ii + dH_vii[2], dH_vii[0] - 1j * dH_vii[1]],
                         [dH_vii[0] + 1j * dH_vii[1], dH_ii - dH_vii[2]]])
                    P_nsi = kpt.projections[a]
                    F_v = np.einsum('nsiv,stij,ntj', F_nsiv, dH_ssii, P_nsi)
                    F_nsiv *= kpt.eps_n[:, np.newaxis, np.newaxis, np.newaxis]
                    dO_ii = self.setups[a].dO_ii
                    F_v -= np.einsum('nsiv,ij,nsj', F_nsiv, dO_ii, P_nsi)
                    F_av[a] += 2 * F_v.real

            self.bd.comm.sum(F_av, 0)

            if self.bd.comm.rank == 0:
                self.kd.comm.sum(F_av, 0)
            return

        F_aniv = self.pt.dict(self.bd.mynbands, derivative=True)
        dH_asp = hamiltonian.dH_asp
        for kpt in self.kpt_u:
            self.pt.derivative(kpt.psit_nG, F_aniv, kpt.q)
            for a, F_niv in F_aniv.items():
                F_niv = F_niv.conj()
                F_niv *= kpt.f_n[:, np.newaxis, np.newaxis]
                dH_ii = unpack(dH_asp[a][kpt.s])
                P_ni = kpt.P_ani[a]
                F_vii = np.dot(np.dot(F_niv.transpose(), P_ni), dH_ii)
                F_niv *= kpt.eps_n[:, np.newaxis, np.newaxis]
                dO_ii = self.setups[a].dO_ii
                F_vii -= np.dot(np.dot(F_niv.transpose(), P_ni), dO_ii)
                F_av[a] += 2 * F_vii.real.trace(0, 1, 2)

            # Hack used in delta-scf calculations:
            if hasattr(kpt, 'c_on'):
                assert self.bd.comm.size == 1
                self.pt.derivative(kpt.psit_nG, F_aniv, kpt.q)  # XXX again
                d_nn = np.zeros((self.bd.mynbands, self.bd.mynbands),
                                dtype=complex)
                for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                    d_nn += ne * np.outer(c_n.conj(), c_n)
                for a, F_niv in F_aniv.items():
                    F_niv = F_niv.conj()
                    dH_ii = unpack(dH_asp[a][kpt.s])
                    Q_ni = np.dot(d_nn, kpt.P_ani[a])
                    F_vii = np.dot(np.dot(F_niv.transpose(), Q_ni), dH_ii)
                    F_niv *= kpt.eps_n[:, np.newaxis, np.newaxis]
                    dO_ii = self.setups[a].dO_ii
                    F_vii -= np.dot(np.dot(F_niv.transpose(), Q_ni), dO_ii)
                    F_av[a] += 2 * F_vii.real.trace(0, 1, 2)

        self.bd.comm.sum(F_av, 0)

        if self.bd.comm.rank == 0:
            self.kd.comm.sum(F_av, 0)

    def estimate_memory(self, mem):
        gridbytes = self.bytes_per_wave_function()
        n = len(self.kpt_u) * self.bd.mynbands
        mem.subnode('Arrays psit_nG', n * gridbytes)
        self.eigensolver.estimate_memory(mem.subnode('Eigensolver'), self)
        ni = sum(dataset.ni for dataset in self.setups) / self.gd.comm.size
        mem.subnode('Projections', n * ni * np.dtype(self.dtype).itemsize)
        self.pt.estimate_memory(mem.subnode('Projectors'))



import numpy as np
import pytest

from gpaw.new.pot_calc import calculate_non_local_potential1
from gpaw.new.xc import create_functional
from gpaw.setup import create_setup
from gpaw.core import UGDesc
from gpaw.xc import XC


@pytest.mark.soc
def test_energy_from_complex_densmat():

    # Set up objects and matrices

    setup = create_setup('Ga')
    grid = UGDesc(cell=[1, 1, 1], size=[9, 9, 9])
    xc = create_functional(XC('LDA', collinear=False), grid)
    soc = True
    err = 1.0e-6
    D_sii = np.zeros((4, setup.ni, setup.ni), complex)
    D_sii[0, 0, 0] = 2

    # Calculate non-local potential energy for some random state

    P_sm = np.array([[0.2 + 0.3j, 0.1 + 0.2j, 0.3 + 0.4j],
                     [0.4 - 0.5j, 0.2 + 0.3j, 0.6 - 0.7j]])

    D_ssmm = np.einsum('si, zj -> szij', P_sm.conj(), P_sm)
    D_sii[:, 1:4, 1:4] = [D_ssmm[0, 0] + D_ssmm[1, 1],
                          D_ssmm[0, 1] + D_ssmm[1, 0],
                          -1j * (D_ssmm[0, 1] - D_ssmm[1, 0]),
                          D_ssmm[0, 0] - D_ssmm[1, 1]]

    _, energies1 = calculate_non_local_potential1(setup, xc, D_sii,
                                                  np.zeros(1), soc)

    assert energies1['kinetic'] == pytest.approx(0.04340694003, abs=err)
    assert energies1['coulomb'] == pytest.approx(-5.5575386716, abs=err)
    assert energies1['zero'] == pytest.approx(-2.432694074696, abs=err)
    assert energies1['xc'] == pytest.approx(1.5938337327, abs=err)

    # Rotate the state 90 degrees around the z-axis (x -> y, y -> -x, z -> z).
    # Assert that this does not change the energies.

    # First rotate the spins
    P_sm = [P_sm[0, :] * (1 - 1j), P_sm[1, :] * (1 + 1j)] / np.sqrt(2)
    # Then rotate the density
    P_sm = np.matmul([[0, 0, 1], [0, 1, 0], [-1, 0, 0]], P_sm.T).T

    D_ssmm = np.einsum('si, zj -> szij', P_sm.conj(), P_sm)
    D_sii[:, 1:4, 1:4] = [D_ssmm[0, 0] + D_ssmm[1, 1],
                          D_ssmm[0, 1] + D_ssmm[1, 0],
                          -1j * (D_ssmm[0, 1] - D_ssmm[1, 0]),
                          D_ssmm[0, 0] - D_ssmm[1, 1]]

    _, energies2 = calculate_non_local_potential1(setup, xc, D_sii,
                                                  np.zeros(1), soc)

    assert energies2['kinetic'] == pytest.approx(energies1['kinetic'], abs=err)
    assert energies2['coulomb'] == pytest.approx(energies1['coulomb'], abs=err)
    assert energies2['zero'] == pytest.approx(energies1['zero'], abs=err)
    assert energies2['xc'] == pytest.approx(energies1['xc'], abs=err)

    # Assert that only the kinetic energy changes when the density
    # matrix is forced to be real

    _, energies3 = calculate_non_local_potential1(setup, xc, D_sii.real,
                                                  np.zeros(1), soc)

    assert energies3['kinetic'] == pytest.approx(0.0446930609623, abs=err)
    assert energies3['coulomb'] == pytest.approx(energies1['coulomb'], abs=err)
    assert energies3['zero'] == pytest.approx(energies1['zero'], abs=err)
    assert energies3['xc'] == pytest.approx(energies1['xc'], abs=err)



"""Generate new release of ASE.

This script does not attempt to import ASE - then it would depend on
which ASE is installed and how - but assumes that it is run from the
ASE root directory."""

import subprocess
import re
import argparse
from time import strftime


def runcmd(cmd, output=False, error_ok=False):
    print('Executing:', cmd)
    try:
        if output:
            txt = subprocess.check_output(cmd, shell=True)
            return txt.decode('utf8')
        else:
            return subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError as err:
        if error_ok:
            print('Failed: {}'.format(err))
            print('Continuing...')
        else:
            raise


bash = runcmd


def py(cmd, output=False):
    return runcmd('python3 {}'.format(cmd))


def py2(cmd, output=False):
    return runcmd('python2 {}'.format(cmd))


def git(cmd, error_ok=False):
    cmd = 'git {}'.format(cmd)
    return runcmd(cmd, output=True, error_ok=error_ok)


versionfile = 'gpaw/__init__.py'


def get_version():
    with open(versionfile) as fd:
        return re.search(r"__version__ = '(\S+)'", fd.read()).group(1)


def main():
    p = argparse.ArgumentParser(description='Generate new release of GPAW.',
                                epilog='Run from the root directory of GPAW.')
    p.add_argument('version', nargs='?',
                   help='new version number')
    p.add_argument('--clean', action='store_true',
                   help='delete release branch and tag')
    p.add_argument('--sign', action='store_true')
    p.add_argument('--wheel', action='store_true')
    args = p.parse_args()

    try:
        current_version = get_version()
    except Exception as err:
        p.error('Cannot get version: {}.  Are you in the root directory?'
                .format(err))

    print('Current version: {}'.format(current_version))

    if not args.version:
        p.print_help()
        raise SystemExit

    version = args.version

    branchname = 'gpaw-{}'.format(version)
    current_version = get_version()

    if args.clean:
        print('Cleaning {}'.format(version))
        git('checkout master')
        git('tag -d {}'.format(version), error_ok=True)
        git('branch -D {}'.format(branchname), error_ok=True)
        git('branch -D {}'.format('web-page'), error_ok=True)
        return

    print('New release: {}'.format(version))

    txt = git('status')
    branch = re.match(r'On branch (\S+)', txt).group(1)
    print('Currently on branch {}'.format(repr(branch)))
    if branch != 'master':
        git('checkout master')

    git('checkout -b {}'.format(branchname))

    majormiddle, minor = version.rsplit('.', 1)
    minor = int(minor)
    nextminor = minor + 1
    next_devel_version = '{}.{}b1'.format(majormiddle, nextminor)

    def update_version(version):
        print('Editing {}: version {}'.format(versionfile, version))
        new_versionline = "__version__ = '{}'\n".format(version)
        lines = []
        ok = False
        with open(versionfile) as fd:
            for line in fd:
                if line.startswith('__version__'):
                    ok = True
                    line = new_versionline
                lines.append(line)
        assert ok
        with open(versionfile, 'w') as fd:
            for line in lines:
                fd.write(line)

    update_version(version)

    releasenotes = 'doc/releasenotes.rst'
    lines = []

    searchtxt = re.escape("""\
Git master branch
=================

:git:`master <>`.
""")

    replacetxt = """\
Git master branch
=================

:git:`master <>`.

* No changes yet


{header}
{underline}

{date}: :git:`{version} <../{version}>`
"""

    date = strftime('%d %B %Y').lstrip('0')
    header = 'Version {}'.format(version)
    underline = '=' * len(header)
    replacetxt = replacetxt.format(header=header, version=version,
                                   underline=underline, date=date)

    print('Editing {}'.format(releasenotes))
    with open(releasenotes) as fd:
        txt = fd.read()
    txt, n = re.subn(searchtxt, replacetxt, txt, re.MULTILINE)
    assert n == 1

    with open(releasenotes, 'w') as fd:
        fd.write(txt)

    searchtxt = """\
News
====
"""

    replacetxt = """\
News
====

* :ref:`GPAW version {version} <releasenotes>` released ({date}).
"""

    replacetxt = replacetxt.format(version=version, date=date)

    frontpage = 'doc/index.rst'
    lines = []
    print('Editing {}'.format(frontpage))
    with open(frontpage) as fd:
        txt = fd.read()
    txt, n = re.subn(searchtxt, replacetxt, txt)
    assert n == 1
    with open(frontpage, 'w') as fd:
        fd.write(txt)

    installdoc = 'doc/install.rst'
    print('Editing {}'.format(installdoc))

    with open(installdoc) as fd:
        txt = fd.read()

    txt, nsub = re.subn(r'gpaw-\d+\.\d+.\d+',
                        'gpaw-{}'.format(version), txt)
    assert nsub > 0
    txt, nsub = re.subn(r'git clone -b \d+\.\d+.\d+',
                        'git clone -b {}'.format(version), txt)
    assert nsub == 1

    with open(installdoc, 'w') as fd:
        fd.write(txt)

    sphinxconf = 'doc/conf.py'
    print('Editing {}'.format(sphinxconf))
    comment = '# This line auto-edited by newrelease script'
    line1 = "dev_version = '{}'  {}\n".format(next_devel_version, comment)
    line2 = "stable_version = '{}'  {}\n".format(version, comment)
    lines = []
    with open(sphinxconf) as fd:
        for line in fd:
            if re.match('dev_version = ', line):
                line = line1
            if re.match('stable_version = ', line):
                line = line2
            lines.append(line)
    with open(sphinxconf, 'w') as fd:
        fd.write(''.join(lines))

    git('add {}'.format(' '.join([versionfile, sphinxconf, installdoc,
                                  frontpage, releasenotes])))
    git('commit -m "GPAW version {}"'.format(version))
    git('tag {0} {1} -m "gpaw-{1}"'
        .format('-s' if args.sign else '',
                version))

    py('setup.py sdist > setup_sdist.log')
    if args.wheel:
        py2('setup.py bdist_wheel > setup_bdist_wheel2.log')
        py('setup.py bdist_wheel > setup_bdist_wheel3.log')
    if args.sign:
        bash('gpg --armor --yes --detach-sign dist/gpaw-{}.tar.gz'
             .format(version))
    git('checkout -b web-page')
    git('branch --set-upstream-to=origin/web-page')
    git('checkout {}'.format(branchname))
    update_version(next_devel_version)
    git('add {}'.format(versionfile))
    git('branch --set-upstream-to=master')
    git('commit -m "bump version number to {}"'.format(next_devel_version))

    print()
    print('Automatic steps done.')
    print()
    print('Now is a good time to:')
    print(' * check the diff')
    print(' * run the tests')
    print(' * verify the web-page build')
    print()
    print('Remaining steps')
    print('===============')
    print('git show {}  # Inspect!'.format(version))
    print('git checkout master')
    print('git merge {}'.format(branchname))
    print('twine upload '
          'dist/gpaw-{v}.tar.gz '
          'dist/gpaw-{v}-py2-none-any.whl '
          'dist/gpaw-{v}-py3-none-any.whl '
          'dist/gpaw-{v}.tar.gz.asc'.format(v=version))
    print('git push --tags origin master  # Assuming your remote is "origin"')
    print('git checkout web-page')
    print('git push --force origin web-page')


if __name__ == '__main__':
    main()



"""This module implements a class for (true) time-dependent density
functional theory calculations.

"""
import time
import warnings
from math import log

import numpy as np

from gpaw.calculator import GPAW
from gpaw.mixer import DummyMixer
from gpaw.preconditioner import Preconditioner
from gpaw.tddft.units import (attosec_to_autime, autime_to_attosec,
                              aufrequency_to_eV)
from gpaw.tddft.utils import MultiBlas
from gpaw.tddft.solvers import create_solver
from gpaw.tddft.propagators import \
    create_propagator, \
    AbsorptionKick
from gpaw.tddft.tdopers import \
    TimeDependentHamiltonian, \
    TimeDependentOverlap, \
    TimeDependentWaveFunctions, \
    TimeDependentDensity, \
    AbsorptionKickHamiltonian
from gpaw.wavefunctions.fd import FD

from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.magneticmomentwriter import MagneticMomentWriter
from gpaw.lcaotddft.restartfilewriter import RestartFileWriter


__all__ = ['TDDFT', 'photoabsorption_spectrum',
           'DipoleMomentWriter', 'MagneticMomentWriter',
           'RestartFileWriter']


# T^-1
# Bad preconditioner
class KineticEnergyPreconditioner:
    def __init__(self, gd, kin, dtype):
        self.preconditioner = Preconditioner(gd, kin, dtype)
        self.preconditioner.allocate()

    def apply(self, kpt, psi, psin):
        for i in range(len(psi)):
            psin[i][:] = self.preconditioner(psi[i], kpt.phase_cd, None, None)


# S^-1
class InverseOverlapPreconditioner:
    """Preconditioner for TDDFT."""
    def __init__(self, overlap):
        self.overlap = overlap

    def apply(self, kpt, psi, psin):
        self.overlap.apply_inverse(psi, psin, kpt)
# ^^^^^^^^^^


class FDTDDFTMode(FD):
    def __call__(self, *args, **kwargs):
        reuse_wfs_method = kwargs.pop('reuse_wfs_method', None)
        assert reuse_wfs_method is None
        return TimeDependentWaveFunctions(self.nn, *args, **kwargs)


class TDDFT(GPAW):
    """Time-dependent density functional theory calculation based on GPAW.

    This class is the core class of the time-dependent density functional
    theory implementation and is the only class which a user has to use.
    """

    def __init__(self, filename: str, *,
                 td_potential: object = None,
                 calculate_energy: bool = True,
                 propagator: dict = None,
                 solver: dict = None,
                 tolerance: float = None,  # deprecated
                 parallel: dict = None,
                 communicator: object = None,
                 txt: str = '-'):
        """
        Parameters
        ----------
        filename
            File containing ground state or time-dependent state to propagate.
        td_potential
            Function class for the time-dependent potential. Must have a method
            ``strength(time)`` which returns the strength of
            the linear potential to each direction as a vector of three floats.
        calculate_energy
            Whether to calculate energy during propagation.
        propagator
            Time propagator for the Kohn-Sham wavefunctions.
        solver
            The iterative linear equations solver for propagator.
        tolerance
            Deprecated. Do not use this, but use solver dictionary instead.
            Tolerance for the linear solver.
        parallel
            Parallelization options
        communicator
            MPI communicator
        txt
            Text output
        """
        # Default values
        if propagator is None:
            propagator = dict(name='SICN')
        if solver is None:
            solver = dict(name='CSCG', tolerance=1e-8)

        assert filename is not None

        # For communicating with observers
        self.action = ''

        self.time = 0.0
        self.kick_strength = np.array([0.0, 0.0, 0.0], dtype=float)
        self.niter = 0
        self.dm_file = None  # XXX remove and use observer instead

        # Parallelization dictionary should default to strided bands
        # but it does not work XXX
        # self.default_parallel = GPAW.default_parallel.copy()
        # self.default_parallel['stridebands'] = True

        self.default_parameters = GPAW.default_parameters.copy()
        self.default_parameters['mixer'] = DummyMixer()
        self.default_parameters['experimental']['reuse_wfs_method'] = None

        # NB: TDDFT restart files contain additional information which
        #     will override the initial settings for time/kick/niter.
        GPAW.__init__(self, filename, parallel=parallel,
                      communicator=communicator, txt=txt)
        if len(self.symmetry.op_scc) > 1:
            raise ValueError('Symmetries are not allowed for TDDFT. '
                             'Run the ground state calculation with '
                             'symmetry={"point_group": False}.')

        assert isinstance(self.wfs, TimeDependentWaveFunctions)
        assert isinstance(self.wfs.overlap, TimeDependentOverlap)

        self.set_positions()

        # Don't be too strict
        self.density.charge_eps = 1e-5

        self.rank = self.wfs.world.rank

        self.calculate_energy = calculate_energy
        if self.hamiltonian.xc.name.startswith('GLLB'):
            self.log('GLLB model potential. Not updating energy.')
            self.calculate_energy = False

        # Time-dependent variables and operators
        self.td_hamiltonian = TimeDependentHamiltonian(self.wfs, self.spos_ac,
                                                       self.hamiltonian,
                                                       td_potential)
        self.td_density = TimeDependentDensity(self)

        # Solver for linear equations
        if isinstance(solver, str):
            solver = dict(name=solver)
        if tolerance is not None:
            warnings.warn(
                "Please specify the solver tolerance using dictionary "
                "solver={'name': name, 'tolerance': tolerance}. "
                "Confirm the used tolerance from the output file. "
                "The old syntax will throw an error in the future.",
                FutureWarning)
            solver.update(tolerance=tolerance)
        self.solver = create_solver(solver)

        # Preconditioner
        # No preconditioner as none good found
        self.preconditioner = None  # TODO! check out SSOR preconditioning
        # self.preconditioner = InverseOverlapPreconditioner(self.overlap)
        # self.preconditioner = KineticEnergyPreconditioner(
        #    self.wfs.gd, self.td_hamiltonian.hamiltonian.kin, complex)

        # Time propagator
        if isinstance(propagator, str):
            propagator = dict(name=propagator)
        self.propagator = create_propagator(propagator)

        self.hpsit = None
        self.eps_tmp = None
        self.mblas = MultiBlas(self.wfs.gd)

        self.tddft_initialized = False

    def tddft_init(self):
        if self.tddft_initialized:
            return

        self.log('')
        self.log('')
        self.log('------------------------------------------')
        self.log('  Time-propagation TDDFT                  ')
        self.log('------------------------------------------')
        self.log('')

        self.log('Charge epsilon:', self.density.charge_eps)

        # Density mixer
        self.td_density.density.set_mixer(DummyMixer())

        # Solver
        self.solver.initialize(self.wfs.gd, self.timer)
        self.log('Solver:', self.solver.todict())

        # Preconditioner
        self.log('Preconditioner:', self.preconditioner)

        # Propagator
        self.propagator.initialize(self.td_density, self.td_hamiltonian,
                                   self.wfs.overlap, self.solver,
                                   self.preconditioner,
                                   self.wfs.gd, self.timer)
        self.log('Propagator:', self.propagator.todict())

        # Parallelization
        wfs = self.wfs
        if self.rank == 0:
            if wfs.kd.comm.size > 1:
                if wfs.nspins == 2:
                    self.log('Parallelization Over Spin')

                if wfs.gd.comm.size > 1:
                    self.log('Using Domain Decomposition: %d x %d x %d' %
                             tuple(wfs.gd.parsize_c))

                if wfs.bd.comm.size > 1:
                    self.log('Parallelization Over bands on %d Processors' %
                             wfs.bd.comm.size)
            self.log('States per processor =', wfs.bd.mynbands)

        # Restarting an FDTD run generates hamiltonian.fdtd_poisson, which
        # now overwrites hamiltonian.poisson
        if hasattr(self.hamiltonian, 'fdtd_poisson'):
            self.hamiltonian.poisson = self.hamiltonian.fdtd_poisson
            self.hamiltonian.poisson.set_grid_descriptor(self.density.finegd)

        # For electrodynamics mode
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.set_density(self.density)
            self.hamiltonian.poisson.print_messages(self.log)
            self.log.flush()

        # Update density and Hamiltonian
        self.propagator.update_time_dependent_operators(self.time)

        # XXX remove dipole moment handling and use observer instead
        self._dm_args0 = (self.density.finegd.integrate(self.density.rhot_g),
                          self.calculate_dipole_moment())

        # Call observers
        self.action = 'init'
        self.call_observers(self.niter)

        self.tddft_initialized = True

    def create_wave_functions(self, mode, *args, **kwargs):
        mode = FDTDDFTMode(mode.nn, mode.interpolation, True)
        GPAW.create_wave_functions(self, mode, *args, **kwargs)

    def read(self, filename):
        reader = GPAW.read(self, filename)
        if 'tddft' in reader:
            r = reader.tddft
            self.time = r.time
            self.niter = r.niter
            self.kick_strength = r.kick_strength

    def _write(self, writer, mode):
        GPAW._write(self, writer, mode)
        w = writer.child('tddft')
        w.write(time=self.time,
                niter=self.niter,
                kick_strength=self.kick_strength)

    def propagate(self, time_step: float, iterations: int,
                  dipole_moment_file: str = None,  # deprecated
                  restart_file: str = None,  # deprecated
                  dump_interval: int = None,  # deprecated
                  ):
        """Propagates wavefunctions.

        Parameters
        ----------
        time_step
            Time step in attoseconds (10^-18 s).
        iterations
            Number of iterations.
        dipole_moment_file
            Deprecated. Do not use this.
            Name of the data file where to the time-dependent dipole
            moment is saved.
        restart_file
            Deprecated. Do not use this.
            Name of the restart file.
        dump_interval
            Deprecated. Do not use this.
            After how many iterations restart data is dumped.
        """
        self.tddft_init()

        if self.propagator.todict()['name'] in ['EFSICN', 'EFSICN_HGH']:
            msg = ("You are using propagator for Ehrenfest dynamics. "
                   "Please use regular propagator.")
            raise RuntimeError(msg)

        def warn_deprecated(parameter, observer):
            warnings.warn(
                f"The {parameter} parameter is deprecated. "
                f"Please use {observer} observer instead. "
                "The old syntax will throw an error in the future.",
                FutureWarning)

        if dipole_moment_file is not None:
            warn_deprecated('dipole_moment_file', 'DipoleMomentWriter')

        if restart_file is not None:
            warn_deprecated('restart_file', 'RestartFileWriter')

        if dump_interval is not None:
            warn_deprecated('dump_interval', 'RestartFileWriter')
        else:
            dump_interval = 100

        if self.rank == 0:
            self.log()
            self.log('Starting time: %7.2f as'
                     % (self.time * autime_to_attosec))
            self.log('Time step:     %7.2f as' % time_step)
            header = """\
                        Simulation     Total         log10     Iterations:
             Time          time        Energy (eV)   Norm      Propagator"""
            self.log()
            self.log(header)

        # Convert to atomic units
        time_step = time_step * attosec_to_autime

        if dipole_moment_file is not None:
            self.initialize_dipole_moment_file(dipole_moment_file)

        # Set these as class properties for use of observers
        self.time_step = time_step
        self.dump_interval = dump_interval  # XXX remove, deprecated
        self.restart_file = restart_file  # XXX remove, deprecated

        niterpropagator = 0
        self.maxiter = self.niter + iterations

        # FDTD requires extra care
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.set_time(self.time)
            self.hamiltonian.poisson.set_time_step(self.time_step)

            # The propagate calculation_mode causes classical part to evolve
            # in time when self.hamiltonian.poisson.solve(...) is called
            self.hamiltonian.poisson.set_calculation_mode('propagate')

            # During each time step, self.hamiltonian.poisson.solve may be
            # called several times (depending on the used propagator).
            # Using the attached observer one ensures that actual propagation
            # takes place only once. This is because the FDTDPoissonSolver
            # changes the calculation_mode from propagate to
            # something else when the propagation is finished.
            self.attach(self.hamiltonian.poisson.set_calculation_mode, 1,
                        'propagate')

        self.timer.start('Propagate')
        while self.niter < self.maxiter:
            norm = self.density.finegd.integrate(self.density.rhot_g)

            # Write dipole moment at every iteration
            if dipole_moment_file is not None:
                if self._dm_args0 is not None:
                    self.update_dipole_moment_file(*self._dm_args0)
                    self._dm_args0 = None
                else:
                    dm = self.calculate_dipole_moment()
                    self.update_dipole_moment_file(norm, dm)

            # Propagate the Kohn-Shame wavefunctions a single timestep
            niterpropagator = self.propagator.propagate(self.time, time_step)
            self.time += time_step
            self.niter += 1

            # print output (energy etc.) every 10th iteration
            if self.niter % 10 == 0:
                self.get_td_energy()

                T = time.localtime()
                if self.rank == 0:
                    iter_text = 'iter: %3d  %02d:%02d:%02d %11.2f' \
                                '   %13.6f %9.1f %10d'
                    self.log(iter_text %
                             (self.niter, T[3], T[4], T[5],
                              self.time * autime_to_attosec,
                              self.Etot * aufrequency_to_eV,
                              log(abs(norm) + 1e-16) / log(10),
                              niterpropagator))

                    self.log.flush()

            # Call registered callback functions
            self.action = 'propagate'
            self.call_observers(self.niter)

            # Write restart data
            if restart_file is not None and self.niter % dump_interval == 0:
                self.write(restart_file, 'all')
                if self.rank == 0:
                    print('Wrote restart file.')
                    print(self.niter, ' iterations done. Current time is ',
                          self.time * autime_to_attosec, ' as.')

        self.timer.stop('Propagate')

        # Write final results and close dipole moment file
        if dipole_moment_file is not None:
            # TODO final iteration is propagated, but nothing is updated
            # norm = self.density.finegd.integrate(self.density.rhot_g)
            # self.finalize_dipole_moment_file(norm)
            self.finalize_dipole_moment_file()

        # Finalize FDTDPoissonSolver
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.finalize_propagation()

        if restart_file is not None:
            self.write(restart_file, 'all')

    def initialize_dipole_moment_file(self, dipole_moment_file):
        if self.rank == 0:
            if self.dm_file is not None and not self.dm_file.closed:
                raise RuntimeError('Dipole moment file is already open')

            if self.time == 0.0:
                mode = 'w'
            else:
                # We probably continue from restart
                mode = 'a'

            self.dm_file = open(dipole_moment_file, mode)

            # If the dipole moment file is empty, add a header
            if self.dm_file.tell() == 0:
                header = '# Kick = [%22.12le, %22.12le, %22.12le]\n' \
                    % (self.kick_strength[0], self.kick_strength[1],
                       self.kick_strength[2])
                header += '# %15s %15s %22s %22s %22s\n' \
                    % ('time', 'norm', 'dmx', 'dmy', 'dmz')
                self.dm_file.write(header)
                self.dm_file.flush()

    def calculate_dipole_moment(self):
        dm = self.density.finegd.calculate_dipole_moment(self.density.rhot_g)
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            dm += self.hamiltonian.poisson.get_classical_dipole_moment()
        return dm

    def update_dipole_moment_file(self, norm, dm):
        if self.rank == 0:
            line = '%20.8lf %20.8le %22.12le %22.12le %22.12le\n' \
                % (self.time, norm, dm[0], dm[1], dm[2])
            self.dm_file.write(line)
            self.dm_file.flush()

    def finalize_dipole_moment_file(self, norm=None):
        if norm is not None:
            dm = self.calculate_dipole_moment()
            self.update_dipole_moment_file(norm, dm)

        if self.rank == 0:
            self.dm_file.close()
            self.dm_file = None

    def update_eigenvalues(self):

        kpt_u = self.wfs.kpt_u
        if self.hpsit is None:
            self.hpsit = self.wfs.gd.zeros(len(kpt_u[0].psit_nG),
                                           dtype=complex)
        if self.eps_tmp is None:
            self.eps_tmp = np.zeros(len(kpt_u[0].eps_n),
                                    dtype=complex)

        # self.Eband = sum_i <psi_i|H|psi_j>
        for kpt in kpt_u:
            self.td_hamiltonian.apply(kpt, kpt.psit_nG, self.hpsit,
                                      calculate_P_ani=False)
            self.mblas.multi_zdotc(self.eps_tmp, kpt.psit_nG,
                                   self.hpsit, len(kpt_u[0].psit_nG))
            self.eps_tmp *= self.wfs.gd.dv
            kpt.eps_n[:] = self.eps_tmp.real

        H = self.td_hamiltonian.hamiltonian

        # Nonlocal
        self.Enlkin = H.xc.get_kinetic_energy_correction()

        # PAW
        e_band = self.wfs.calculate_band_energy()
        self.Ekin = H.e_kinetic0 + e_band + self.Enlkin
        self.e_coulomb = H.e_coulomb
        self.Eext = H.e_external
        self.Ebar = H.e_zero
        self.Exc = H.e_xc
        self.Etot = self.Ekin + self.e_coulomb + self.Ebar + self.Exc

    def get_td_energy(self):
        """Calculate the time-dependent total energy"""
        self.tddft_init()

        if not self.calculate_energy:
            self.Etot = 0.0
            return 0.0

        self.wfs.overlap.update(self.wfs)
        self.td_density.update()
        self.td_hamiltonian.update(self.td_density.get_density(),
                                   self.time)
        self.update_eigenvalues()

        return self.Etot

    def set_absorbing_boundary(self, absorbing_boundary):
        self.td_hamiltonian.set_absorbing_boundary(absorbing_boundary)

    # exp(ip.r) psi
    def absorption_kick(self, kick_strength):
        """Delta absorption kick for photoabsorption spectrum.

        Parameters
        ----------
        kick_strength
            Strength of the kick in atomic units
        """
        self.tddft_init()

        if self.rank == 0:
            self.log('Delta kick =', kick_strength)

        self.kick_strength = np.array(kick_strength)

        abs_kick_hamiltonian = AbsorptionKickHamiltonian(
            self.wfs, self.spos_ac,
            np.array(kick_strength, float))
        abs_kick = AbsorptionKick(self.wfs, abs_kick_hamiltonian,
                                  self.wfs.overlap, self.solver,
                                  self.preconditioner, self.wfs.gd, self.timer)
        abs_kick.kick()

        # Kick the classical part, if it is present
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.set_kick(kick=self.kick_strength)

        # Update density and Hamiltonian
        self.propagator.update_time_dependent_operators(self.time)

        # Call observers after kick
        self.action = 'kick'
        self.call_observers(self.niter)



from typing import NamedTuple, Dict, List

import numpy as np

from gpaw.mpi import broadcast
from gpaw.utilities import (pack_atomic_matrices, unpack_atomic_matrices,
                            unpack2, unpack, packed_index)


class PAWThings(NamedTuple):
    VC_aii: Dict[int, np.ndarray]
    VV_aii: Dict[int, np.ndarray]  # distributed
    Delta_aiiL: List[np.ndarray]


def calculate_paw_stuff(wfs, dens) -> List[PAWThings]:
    D_asp = dens.D_asp
    comm = D_asp.partition.comm
    if comm.size != wfs.world.size:
        D_sP = pack_atomic_matrices(D_asp)
        D_sP = broadcast(D_sP if comm.rank == 0 else None, comm=comm)
        D_asp = unpack_atomic_matrices(D_sP, wfs.setups)
        rank_a = np.linspace(0, wfs.world.size, len(wfs.setups),
                             endpoint=False).astype(int)
        D_asp = {a: D_sp for a, D_sp in D_asp.items()
                 if rank_a[a] == wfs.world.rank}

    VV_saii: List[Dict[int, np.ndarray]] = [{} for s in range(dens.nspins)]
    for a, D_sp in D_asp.items():
        data = wfs.setups[a]
        for VV_aii, D_p in zip(VV_saii, D_sp):
            D_ii = unpack2(D_p) * (dens.nspins / 2)
            VV_ii = pawexxvv(data.M_pp, D_ii)
            VV_aii[a] = VV_ii

    Delta_aiiL = []
    VC_aii = {}
    for a, data in enumerate(wfs.setups):
        Delta_aiiL.append(data.Delta_iiL)
        VC_aii[a] = unpack(data.X_p)

    return [PAWThings(VC_aii, VV_aii, Delta_aiiL)
            for VV_aii in VV_saii]


def pawexxvv(M_pp, D_ii):
    """PAW correction for valence-valence EXX energy."""
    ni = len(D_ii)
    V_ii = np.empty((ni, ni))
    for i1 in range(ni):
        for i2 in range(ni):
            V = 0.0
            for i3 in range(ni):
                p13 = packed_index(i1, i3, ni)
                for i4 in range(ni):
                    p24 = packed_index(i2, i4, ni)
                    V += M_pp[p13, p24] * D_ii[i3, i4]
            V_ii[i1, i2] = V
    return V_ii



from myqueue.workflow import run
from gpaw.test.big.test_systems.create import create_test_systems
from gpaw import GPAW, PW, MixerFull
from ase.optimize import BFGS


CORES = [1, 4, 8, 16, 14, 40, 48, 56, 72, 96, 120, 168]


def workflow():
    for name, (atoms, params) in create_test_systems().items():
        if name == 'biimtf':
            continue
        cores = len(atoms)**2 / 10
        # Find best match:
        _, cores = min((abs(cores - c), c) for c in CORES)
        run(function=calculate, args=[name, atoms, params],
            name=name,
            cores=cores,
            tmax='1d')
        # run(function=relax, args=[name, atoms, params],
        #     name=name + '-R',
        #     cores=cores,
        #     tmax='1d')


def calculate(name, atoms, params):
    """Do one-shot energy calculation."""
    atoms.calc = GPAW(**params,
                      mixer=MixerFull(),
                      txt=name + '.txt')
    atoms.get_potential_energy()


def relax(name, atoms, params):
    """Relax with PW-mode."""
    kwargs = {k: v for k, v in params.items() if k in ['charge', 'spinpol']}
    atoms.calc = GPAW(
        xc='PBE',
        mode=PW(600),
        kpts=dict(density=4.0),
        mixer=MixerFull(),
        txt=name + '-R.txt',
        **kwargs)
    BFGS(atoms,
         logfile=name + '-R.log',
         trajectory=name + '-R.traj').run(fmax=0.02)



import pytest
from ase import Atom, Atoms

from gpaw import GPAW
from gpaw.analyse.overlap import Overlap
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.test import equal

txt = '-'
txt = None
load = True
load = False
xc = 'LDA'

# run

R = 0.7  # approx. experimental bond length
a = 4.0
c = 5.0


@pytest.mark.lrtddft
def test_lrtddft_placzek_profeta_albrecht(in_tmp_dir):
    from ase.vibrations.albrecht import Albrecht
    from ase.vibrations.placzek import Placzek, Profeta
    from ase.vibrations.resonant_raman import ResonantRamanCalculator

    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    name = exname = 'rraman'
    exkwargs = {'restrict': {'eps': 0.0, 'jend': 3}}

    calc = GPAW(
        mode='fd',
        xc=xc, nbands=7,
        convergence={'bands': 3},
        spinpol=False,
        # eigensolver='rmm-diis',
        symmetry={'point_group': False},
        txt=txt)
    H2.calc = calc
    # H2.get_potential_energy()

    rr = ResonantRamanCalculator(
        H2, KSSingles, name=name, exname=exname,
        exkwargs=exkwargs,
        # XXX full does not work in parallel due to boxes
        # on different nodes
        # overlap=lambda x, y: Overlap(x).full(y)[0],
        overlap=lambda x, y: Overlap(x).pseudo(y)[0],
        txt=txt)
    rr.run()

    # check

    # Different Placzeck implementations should agree

    om = 5
    pz = Placzek(H2, KSSingles, name=name, exname=exname, txt=txt)
    pzi = pz.get_absolute_intensities(omega=om)[-1]

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 approximation='Placzek', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]
    equal(pzi, pri, 0.1)

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='Placzek', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]
    equal(pzi, pri, 0.1)

    """Albrecht and Placzek are approximately equal"""

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pzi, ali, 1.5)

    """Albrecht A and P-P are approximately equal"""

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='P-P', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht A', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pri, ali, 3)

    """Albrecht B+C and Profeta are approximately equal"""

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='Profeta', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht BC', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pri, ali, 3)



import pytest
from ase import Atom, Atoms
from gpaw import GPAW, FermiDirac


def test_spin_spinpol():
    a = 4.0
    n = 16
    hydrogen = Atoms([Atom('H')], cell=(a, a, a), pbc=True)
    hydrogen.center()
    hydrogen.calc = GPAW(
        mode='fd',
        gpts=(n, n, n),
        nbands=1,
        convergence={'energy': 1e-5},
        occupations=FermiDirac(0.0))
    e1 = hydrogen.get_potential_energy()

    hydrogen.calc = hydrogen.calc.new(hund=True)
    e2 = hydrogen.get_potential_energy()

    de = e1 - e2
    print(de)
    assert de == pytest.approx(0.7871, abs=1.e-4)

    energy_tolerance = 0.0006
    assert e1 == pytest.approx(-0.499854, abs=energy_tolerance)
    assert e2 == pytest.approx(-1.287, abs=energy_tolerance)



import os
import sys
import time

import numpy as np

import ase.db
from ase.utils import opencew
from ase.calculators.calculator import kpts2mp
from ase.io import Trajectory
from ase.calculators.aims import Aims
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection

collection = Collection()

if len(sys.argv) == 1:
    names = collection.names
else:
    names = [sys.argv[1]]

c = ase.db.connect('dcdft_aims.db')

# select the basis set
basis = 'light'
# basis = 'tight'
# basis = 'really_tight'
# basis = 'tier2'

kptdensity = 16.0  # this is converged
kptdensity = 6.0  # just for testing
width = 0.01

basis_threshold = 0.00001
relativistic = 'none'
relativistic = 1.e-12
relativistic = 'scalar'

sc_accuracy_rho = 1.e-4
sc_accuracy_eev = 5.e-3

if relativistic == 'none':
    linspace = (0.92, 1.08, 7)  # eos numpy's linspace
else:
    linspace = (0.98, 1.02, 5)  # eos numpy's linspace
linspacestr = ''.join([str(t) + 'x' for t in linspace])[:-1]

code = 'aims' + '-' + basis + '_e' + linspacestr
code = code + '_k' + str(kptdensity) + '_w' + str(width)
code = code + '_t' + str(basis_threshold) + '_r' + str(relativistic)

collection = Collection()

for name in names:
    # save all steps in one traj file in addition to the database
    # we should only used the database c.reserve, but here
    # traj file is used as another lock ...
    fd = opencew(name + '_' + code + '.traj')
    if fd is None:
        continue
    traj = Trajectory(name + '_' + code + '.traj', 'w')
    atoms = collection[name]
    cell = atoms.get_cell()
    kpts = tuple(kpts2mp(atoms, kptdensity, even=True))
    kwargs = {}
    if relativistic == 'scalar':
        kwargs.update({'relativistic': ['atomic_zora', relativistic]})
    elif relativistic == 'none':
        kwargs.update({'relativistic': 'none'})
    else:  # e.g. 1.0e-12
        kwargs.update({'relativistic': ['zora', 'scalar', relativistic]})
    if atoms.get_initial_magnetic_moments().any():  # spin-polarization
        magmom = atoms.get_initial_magnetic_moments().sum() / len(atoms)
        kwargs.update({'spin': 'collinear'})
    # convergence problems for tier2
    charge_mix_param = 0.01
    basis_threshold = 0.00001
    if basis in ['tier2']:
        if name in ['Cr', 'Fe'] and relativistic == 'none':
            basis_threshold = 0.00005
            sc_accuracy_rho = 2.5e-3
            sc_accuracy_eev = 5.e-3
        if name in ['Mn']:
            charge_mix_param = 0.01
            basis_threshold = 0.00005
            sc_accuracy_rho = 2.5e-3
            sc_accuracy_eev = 5.e-3
            if relativistic == 'none':
                sc_accuracy_rho = 3.0e-3
    # loop over EOS linspace
    for n, x in enumerate(np.linspace(linspace[0], linspace[1], linspace[2])):
        id = c.reserve(name=name, basis=basis, linspacestr=linspacestr,
                       kptdensity=kptdensity, width=width,
                       basis_threshold=basis_threshold,
                       relativistic=relativistic,
                       x=x)
        if id is None:
            continue
        # perform EOS step
        atoms.set_cell(cell * x, scale_atoms=True)
        # set calculator
        atoms.calc = Aims(
            label=name + '_' + code + '_' + str(n),
            species_dir=os.path.join(os.environ['AIMS_SPECIES_DIR'], basis),
            xc='PBE',
            kpts=kpts,
            KS_method='elpa',
            sc_accuracy_rho=sc_accuracy_rho,
            sc_accuracy_eev=sc_accuracy_eev,
            occupation_type=['gaussian', width],
            override_relativity=True,
            override_illconditioning=True,
            basis_threshold=basis_threshold,
            charge_mix_param=charge_mix_param,
            sc_iter_limit=9000)
        atoms.calc.set(**kwargs)  # remaining calc keywords
        t = time.time()
        atoms.get_potential_energy()
        c.write(atoms,
                name=name, basis=basis, linspacestr=linspacestr,
                kptdensity=kptdensity, width=width,
                basis_threshold=basis_threshold,
                relativistic=relativistic,
                x=x,
                time=time.time() - t)
        traj.write(atoms)
        del c[id]



import os
import sys
import time

import numpy as np
import ase
from ase import __version__ as ase_version
from ase.utils import search_current_git_hash, IOContext

import gpaw
import _gpaw
from gpaw.utilities.memory import maxrss


class GPAWLogger:
    """Class for handling all text output."""
    def __init__(self, world):
        self.world = world

        self.verbose = False
        self._fd = None
        self.oldfd = 42
        self.iocontext = IOContext()

    @property
    def fd(self):
        return self._fd

    @fd.setter
    def fd(self, fd):
        """Set the stream for text output.

        If `txt` is not a stream-object, then it must be one of:

        * None:  Throw output away.
        * '-':  Use stdout (``sys.stdout``) on master, elsewhere throw away.
        * A filename:  Open a new file on master, elsewhere throw away.
        """
        if fd == self.oldfd:
            return
        self.oldfd = fd
        self._fd = self.iocontext.openfile(fd, self.world)
        self.header()

    def __call__(self, *args, **kwargs):
        flush = kwargs.pop('flush', False)
        print(*args, file=self._fd, **kwargs)
        if flush:
            self._fd.flush()

    def flush(self):
        self._fd.flush()

    def header(self):
        self()
        self('  ___ ___ ___ _ _ _  ')
        self(' |   |   |_  | | | | ')
        self(' | | | | | . | | | | ')
        self(' |__ |  _|___|_____| ', gpaw.__version__)
        self(' |___|_|             ')
        self()

        write_header(self, self.world)

    def print_dict(self, dct, sep='  '):
        options = np.get_printoptions()
        try:
            np.set_printoptions(threshold=4, linewidth=50)
            for key, value in sorted(dct.items()):
                if hasattr(value, 'todict'):
                    value = value.todict()
                if isinstance(value, dict):
                    sep = ',\n     ' + ' ' * len(key)
                    keys = sorted(value, key=lambda k: (str(type(k)), k))
                    s = sep.join('{0}: {1}'.format(k, value[k]) for k in keys)
                    self('  {0}: {{{1}}}'.format(key, s))
                elif hasattr(value, '__len__'):
                    value = np.asarray(value)
                    sep = ',\n    ' + ' ' * len(key)
                    s = sep.join(str(value).splitlines())
                    self('  {0}: {1}'.format(key, s))
                else:
                    self('  {0}: {1}'.format(key, value))
        finally:
            np.set_printoptions(**options)

    def __del__(self):
        """Destructor:  Write timing output before closing."""
        self.close()

    def close(self):
        if gpaw.dry_run or self._fd.closed:
            return

        try:
            mr = maxrss()
        except (LookupError, TypeError, NameError, AttributeError):
            # Thing can get weird during interpreter shutdown ...
            mr = 0

        if mr > 0:
            if mr < 1024**3:
                self('Memory usage: %.2f MiB' % (mr / 1024**2))
            else:
                self('Memory usage: %.2f GiB' % (mr / 1024**3))

        self('Date: ' + time.asctime())

        self.iocontext.close()


def write_header(log, world):
    # We use os.uname() here bacause platform.uname() starts a subprocess,
    # which MPI may not like!
    # This might not work on Windows.  We will see ...
    nodename, machine = os.uname()[1::3]

    log('User:  ', os.getenv('USER', '???') + '@' + nodename)
    log('Date:  ', time.asctime())
    log('Arch:  ', machine)
    log('Pid:   ', os.getpid())
    log('CWD:   ', os.getcwd())
    log('Python: {0}.{1}.{2}'.format(*sys.version_info[:3]))
    # GPAW
    line = os.path.dirname(gpaw.__file__)
    githash = search_current_git_hash(gpaw, world)
    if githash is not None:
        line += ' ({:.10})'.format(githash)
    log('gpaw:  ', line)

    # Find C-code:
    c = getattr(_gpaw, '__file__', None)
    if not c:
        c = sys.executable
    line = os.path.normpath(c)
    if hasattr(_gpaw, 'githash'):
        line += ' ({:.10})'.format(_gpaw.githash())
    log('_gpaw: ', cut(line))

    # ASE
    line = '%s (version %s' % (os.path.dirname(ase.__file__), ase_version)
    githash = search_current_git_hash(ase, world)
    if githash is not None:
        line += '-{:.10}'.format(githash)
    line += ')'
    log('ase:   ', line)

    log('numpy:  %s (version %s)' %
        (os.path.dirname(np.__file__), np.version.version))
    import scipy as sp
    log('scipy:  %s (version %s)' %
        (os.path.dirname(sp.__file__), sp.version.version))
    # Explicitly deleting SciPy seems to remove garbage collection
    # problem of unknown cause
    del sp
    log('libxc: ', getattr(_gpaw, 'libxc_version', '2.x.y'))
    log('units:  Angstrom and eV')
    log('cores:', world.size)
    log('OpenMP:', _gpaw.have_openmp)
    log('OMP_NUM_THREADS:', os.environ['OMP_NUM_THREADS'])

    if gpaw.debug:
        log('DEBUG-MODE: true')

    log()


def cut(s, indent='        '):
    if len(s) + len(indent) < 80:
        return s
    s1, s2 = s.rsplit('/', 1)
    return s1 + '/\n' + indent + s2


def indent(s, level=1, tab='  '):
    return ''.join(tab * level + l for l in s.splitlines(True))



import numpy as np
from ase import Atom, Atoms

import gpaw.mpi as mpi
from gpaw import GPAW, FermiDirac, PoissonSolver
from gpaw.test import equal
from gpaw.utilities.dos import RawLDOS, raw_orbital_LDOS, raw_wignerseitz_LDOS


def test_utilities_ldos(in_tmp_dir, gpaw_new):
    comms = [mpi.world.new_communicator(np.array([r]))
             for r in range(mpi.size)]
    comm = comms[mpi.rank]

    Hnospin = Atoms([Atom('H')], cell=[5, 5, 5], pbc=False)
    Hspin = Atoms([Atom('H', magmom=1)], cell=[5, 5, 5], pbc=False)
    LiH = Atoms([Atom('Li', [.0, .0, .41]),
                 Atom('H', [.0, .0, -1.23])])
    Hnospin.center()
    Hspin.center()
    LiH.center(vacuum=3.0)

    # This is needed for the Wigner-Seitz test to give
    # architecture-independent results:
    LiH.translate(0.003234)

    calc = GPAW(mode='fd', gpts=(24, 24, 24), communicator=comm)
    Hnospin.calc = calc
    e_Hnospin = Hnospin.get_potential_energy()
    energies, sweight = raw_orbital_LDOS(calc, a=0, spin=0, angular='s')
    energies, pdfweight = raw_orbital_LDOS(calc, a=0, spin=0, angular='pdf')

    calc = GPAW(mode='fd',
                gpts=(24, 24, 24),
                occupations=FermiDirac(width=0, fixmagmom=True),
                poissonsolver=PoissonSolver('fd'),
                hund=True,
                communicator=comm)
    Hspin.calc = calc
    e_Hspin = Hspin.get_potential_energy()
    energies, sweight_spin = raw_orbital_LDOS(calc, a=0, spin=0, angular='s')

    calc = GPAW(mode='fd', gpts=(32, 32, 40), nbands=2,
                poissonsolver=PoissonSolver('fd'),
                communicator=comm)
    LiH.calc = calc
    e_LiH = LiH.get_potential_energy()
    energies, Li_orbitalweight = raw_orbital_LDOS(calc, a=0, spin=0,
                                                  angular=None)
    energies, H_orbitalweight = raw_orbital_LDOS(calc, a=1, spin=0,
                                                 angular=None)
    energies, Li_wzweight = raw_wignerseitz_LDOS(calc, a=0, spin=0)
    energies, H_wzweight = raw_wignerseitz_LDOS(calc, a=1, spin=0)

    if not gpaw_new:
        n_a = calc.get_wigner_seitz_densities(spin=0)
        print(n_a)
        equal(n_a.sum(), 0.0, 1e-5)
        equal(n_a[1], 0.737, 0.001)

    print(sweight, pdfweight)
    print(sweight_spin)
    print(Li_wzweight)
    print(H_wzweight)

    equal(sweight[0], 1.0, 0.06)
    equal(pdfweight[0], 0.0, 0.0001)
    equal(sweight_spin[0], 1.14, 0.06)
    assert ((Li_wzweight - [.13, 0.93]).round(2) == 0).all()
    assert ((H_wzweight - [0.87, 0.07]).round(2) == 0).all()
    assert ((Li_wzweight + H_wzweight).round(5) == 1).all()

    print(Li_orbitalweight)
    print(H_orbitalweight)
    #               HOMO    s   py  pz  px  *s
    Li_orbitalweight[0] -= [.5, .0, .6, .0, .0]
    H_orbitalweight[0] -= [.7, .0, .0, .0, .0]

    #              LUMO       s  py   pz  px  *s
    Li_orbitalweight[1] -= [1.0, .0, 0.9, .0, .0]
    H_orbitalweight[1] -= [0.1, .0, 0.0, .0, .0]

    assert not Li_orbitalweight.round(1).any()
    assert not H_orbitalweight.round(1).any()

    ldos = RawLDOS(calc)
    fname = 'ldbe'
    ldos.by_element_to_file(fname + '.dat', shift=False)
    ldos.by_element_to_file(fname + '_2.0.dat', 2.0, shift=False)
    ldos.by_element_to_file(fname + '_indx0.dat', indices=[0])
    # the hydrogen entries are missing for index 0 only
    if mpi.world.rank == 0:
        assert (np.loadtxt(fname + '_indx0.dat').shape[1] + 3 ==
                np.loadtxt(fname + '.dat').shape[1])

    energy_tolerance = 0.001
    equal(e_Hnospin, 0.153991, energy_tolerance)
    equal(e_Hspin, -0.782309, energy_tolerance)
    equal(e_LiH, -3.74582, energy_tolerance)



import pytest
from ase import Atoms

from gpaw import GPAW, LCAO
from gpaw.mixer import FFTMixer
from gpaw.test import equal


@pytest.mark.later
def test_lcao_fftmixer():
    bulk = Atoms('Li', pbc=True,
                 cell=[2.6, 2.6, 2.6])
    k = 4
    bulk.calc = GPAW(mode=LCAO(),
                     kpts=(k, k, k),
                     mixer=FFTMixer())
    e = bulk.get_potential_energy()
    equal(e, -1.710364, 1e-4)



from ase.units import Bohr

import gpaw.mpi as mpi
from gpaw.lrtddft.spectrum import Writer

ds_prefactor = {
    'Ang': Bohr ** 2,
    'a.u.': 1.0,
    'Mb': Bohr ** 2 * 100}


class PESpectrum(Writer):

    def __init__(self,
                 enlist,
                 folding='Gauss',
                 width=0.08):  # Gauss/Lorentz width
        Writer.__init__(self, folding, width)
        self.title = 'Photo emission spectrum'
        self.fields = 'Binding energy [eV]   '
        if folding is None:
            self.fields += 'Spectroscopic factor'
        else:
            self.fields += 'Folded spectroscopic factor'

        self.energies = enlist[0]
        self.values = []
        for val in enlist[1]:
            self.values.append([val])


class BasePES:

    def save_folded_pes(self,
                        filename=None,
                        width=0.08,  # Gauss/Lorentz width
                        emin=None,
                        emax=None,
                        de=None,
                        folding='Gauss',
                        comment=None):

        ew = self.get_energies_and_weights()
        if mpi.rank == 0:
            sp = PESpectrum(ew, folding, width)
            sp.write(filename, emin, emax, de, comment)

    def get_energies_and_weights(self):
        if self.be is None or self.f is None:
            self._calculate()

        return self.be, self.f

    def set_first_peak_energy(self, energy):
        self.first_peak_energy = energy



from ase import Atoms

from gpaw.calculator import GPAW
from gpaw.typing import Array3D


class WannierFunctions:
    def __init__(self,
                 atoms: Atoms,
                 centers,
                 value,
                 U_knn,
                 n1=0,
                 spin=0):
        self.atoms = atoms
        self.centers = centers
        self.U_knn = U_knn
        self.value = value
        self.n1 = 0
        self.spin = spin

    def centers_as_atoms(self):
        return self.atoms + Atoms(f'X{len(self.centers)}', self.centers)

    def get_function(self, calc: GPAW, n: int) -> Array3D:
        wf = calc.wfs.gd.zeros()
        for m, u in enumerate(self.U_knn[0][:, n]):
            wf += u * calc.wfs.get_wave_function_array(n=self.n1 + m,
                                                       s=self.spin,
                                                       k=0)
        return wf



import re

import pytest
import numpy as np
from ase.build import molecule
from ase.units import Hartree

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.test import equal
from gpaw.gauss import Gauss
from gpaw.lrtddft import LrTDDFT, photoabsorption_spectrum
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.lrtddft
@pytest.mark.slow
@pytest.mark.skipif(world.size > 1, reason='test is serial')
def test_lrtddft_3(in_tmp_dir):
    from io import StringIO

    txt = None
    N2 = molecule('N2', vacuum=2.0)

    calc = GPAW(mode='fd',
                h=0.25,
                nbands=-2,
                spinpol=True,
                xc='LDA',
                txt=txt)

    N2.calc = calc
    _ = N2.get_potential_energy()
    calc.write('N2_wfs.gpw', 'all')

    # selections
    for obj in [KSSingles, LrTDDFT]:
        # selection using state numbers
        el = obj(restrict={'istart': 3, 'jend': 6}, txt=txt)
        el.calculate(N2)
        if hasattr(obj, 'diagonalize'):
            el.diagonalize()

        assert len(el) == 8
        # selection using an energy range
        el = obj(restrict={'energy_range': 8}, txt=txt)
        el.calculate(N2)
        if hasattr(obj, 'diagonalize'):
            el.diagonalize()

        assert len(el) == 4
        el = obj(restrict={'energy_range': 11.5}, txt=txt)
        el.calculate(N2)

        if hasattr(obj, 'diagonalize'):
            el.diagonalize()

        # Used to be == 18, but we lowered vacuum so test runs fast
        # and now it's:
        assert len(el) == 16

        if hasattr(obj, 'diagonalize'):
            el.diagonalize(restrict={'energy_range': 8})
            assert len(el) == 4

    lr = LrTDDFT(calc, nspins=2)
    lr.write('lrtddft3.dat.gz')
    lr.diagonalize()

    # This is done to test if writing and reading again yields the same result
    lr2 = LrTDDFT.read('lrtddft3.dat.gz')
    lr2.diagonalize()

    # Post processing

    Epeak = 19.5  # The peak we want to investigate (this is alone)
    Elist = np.asarray([lrsingle.get_energy() * Hartree
                        for lrsingle in lr])
    n = np.argmin(np.abs(Elist - Epeak))  # index of the peak

    E = lr[n].get_energy() * Hartree
    osz = lr[n].get_oscillator_strength()
    print('Original object        :', E, osz[0])

    # Test the output of analyse
    sio = StringIO()
    lr.analyse(n, out=sio)
    s = sio.getvalue()

    match = re.findall(
        r'%i: E=([0-9]*\.[0-9]*) eV, f=([0-9]*\.[0-9]*)*' % n, s)
    Eanalyse = float(match[0][0])
    oszanalyse = float(match[0][1])
    print('From analyse           :', Eanalyse, oszanalyse)
    equal(E, Eanalyse, 1e-3)            # Written precision in analyse
    equal(osz[0], oszanalyse, 1e-3)

    E2 = lr2[n].get_energy() * Hartree
    osz2 = lr2[n].get_oscillator_strength()
    print('Written and read object:', E2, osz2[0])

    # Compare values of original and written/read objects
    equal(E, E2, 1e-4)
    for i in range(len(osz)):
        equal(osz[i], osz2[i], 1.7e-4)

    width = 0.05
    photoabsorption_spectrum(lr,
                             spectrum_file='lrtddft3-spectrum.dat',
                             width=width)
    # We need to be able to check the heights in the spectrum
    weight = Gauss(width).get(0)

    spectrum = np.loadtxt('lrtddft3-spectrum.dat', usecols=(0, 1))
    idx = (spectrum[:, 0] >= E - 0.1) & (spectrum[:, 0] <= E + 0.1)
    peak = np.argmax(spectrum[idx, 1]) + np.nonzero(idx)[0][0]
    Espec = spectrum[peak, 0]
    oszspec = spectrum[peak, 1] / weight

    print('Values from spectrum   :', Espec, oszspec)
    # Compare calculated values with values written to file
    equal(E, Espec, 1e-2)           # The spectrum has a low sampling
    equal(osz[0], oszspec, 1e-2)



import pickle

import numpy as np
from ase.utils import gcd
from scipy.interpolate import InterpolatedUnivariateSpline

from gpaw import GPAW
from gpaw.spinorbit import soc_eigenstates
from gpaw.kpt_descriptor import to1bz


class GWBands:
    """This class defines the GW_bands properties"""

    def __init__(self,
                 calc=None,
                 comm=None,
                 gw_file=None,
                 kpoints=None,
                 bandrange=None):

        self.calc = GPAW(calc, communicator=comm, txt=None)
        if gw_file is not None:
            self.gw_file = pickle.load(open(gw_file, 'rb'), encoding='bytes')
        self.kpoints = kpoints
        if bandrange is None:
            self.bandrange = np.arange(self.calc.get_number_of_bands())
        else:
            self.bandrange = bandrange

        self.gd = self.calc.wfs.gd.new_descriptor()
        self.kd = self.calc.wfs.kd

        self.acell_cv = self.gd.cell_cv
        self.bcell_cv = 2 * np.pi * self.gd.icell_cv

    def find_k_along_path(self, plot_BZ=True):
        """Finds the k-points along the bandpath present in the
           original calculation"""
        kd = self.kd
        acell_cv = self.acell_cv
        bcell_cv = self.bcell_cv
        kpoints = self.kpoints

        if plot_BZ:
            """Plotting the points in the Brillouin Zone"""
            kp_1bz = to1bz(kd.bzk_kc, acell_cv)

            bzk_kcv = np.dot(kd.bzk_kc, bcell_cv)
            kp_1bz_v = np.dot(kp_1bz, bcell_cv)

            import matplotlib.pyplot as plt
            plt.plot(bzk_kcv[:, 0], bzk_kcv[:, 1], 'xg')
            plt.plot(kp_1bz_v[:, 0], kp_1bz_v[:, 1], 'ob')
            for ik in range(1, len(kpoints)):
                kpoint1_v = np.dot(kpoints[ik], bcell_cv)
                kpoint2_v = np.dot(kpoints[ik - 1], bcell_cv)
                plt.plot([kpoint1_v[0], kpoint2_v[0]], [kpoint1_v[1],
                                                        kpoint2_v[1]], '--vr')

        """Finding the points along given directions"""
        print('Finding the kpoints along the path')
        N_c = kd.N_c
        wpts_xc = kpoints

        x_x = []
        k_xc = []
        k_x = []
        x = 0.
        X = []
        for nwpt in range(1, len(wpts_xc)):
            X.append(x)
            to_c = wpts_xc[nwpt]
            from_c = wpts_xc[nwpt - 1]
            vec_c = to_c - from_c
            print('From ', from_c, ' to ', to_c)
            Nv_c = (vec_c * N_c).round().astype(int)
            Nv = abs(gcd(gcd(Nv_c[0], Nv_c[1]), Nv_c[2]))
            print(Nv, ' points found')
            dv_c = vec_c / Nv
            dv_v = np.dot(dv_c, bcell_cv)
            dx = np.linalg.norm(dv_v)
            if nwpt == len(wpts_xc) - 1:
                # X.append(Nv * dx)
                Nv += 1
            for n in range(Nv):
                k_c = from_c + n * dv_c
                bzk_c = to1bz(np.array([k_c]), acell_cv)[0]
                ikpt = kd.where_is_q(bzk_c, kd.bzk_kc)
                x_x.append(x)
                k_xc.append(k_c)
                k_x.append(ikpt)
                x += dx
        X.append(x_x[-1])
        if plot_BZ is True:
            for ik in range(len(k_xc)):
                ktemp_xcv = np.dot(k_xc[ik], bcell_cv)
                plt.plot(ktemp_xcv[0], ktemp_xcv[1], 'xr', markersize=10)
            plt.show()

        return x_x, k_xc, k_x, X

    def get_dft_eigenvalues(self):
        Nk = len(self.calc.get_ibz_k_points())
        bands = np.arange(self.bandrange[0], self.bandrange[-1])
        e_kn = np.array([self.calc.get_eigenvalues(kpt=k)[bands]
                         for k in range(Nk)])
        return e_kn

    def get_vacuum_level(self, plot_pot=False):
        """Finds the vacuum level through Hartree potential"""
        vHt_g = self.calc.get_electrostatic_potential()
        vHt_z = np.mean(np.mean(vHt_g, axis=0), axis=0)

        if plot_pot:
            import matplotlib.pyplot as plt
            plt.plot(vHt_z)
            plt.show()
        return vHt_z[0]

    def get_spinorbit_corrections(self, return_spin=True, return_wfs=False,
                                  bands=None, dft=False,
                                  eig_file=None):
        """Gets the spinorbit corrections to the eigenvalues"""
        calc = self.calc
        bandrange = self.bandrange

        if not dft:
            try:
                e_kn = self.gw_file['qp'][0]
            except KeyError:
                e_kn = self.gw_file[b'qp'][0]
        else:
            if eig_file is not None:
                e_kn = pickle.load(open(eig_file))[0]
            else:
                e_kn = self.get_dft_eigenvalues()[
                    :, bandrange[0]:bandrange[-1] + 1]

        # this will fail - please write a test!
        soc = soc_eigenstates(
            calc,
            n1=bandrange[0],
            n2=bandrange[-1] + 1,
            eigenvalues=e_kn[np.newaxis])
        eSO_nk = soc.eigenvalues().T
        e_kn = eSO_nk.T
        return e_kn

    def get_gw_bands(self, nk_Int=50, interpolate=False, SO=False,
                     dft=False, eig_file=None, vac=False):
        """Getting Eigenvalues along the path"""
        kd = self.kd
        if SO:
            e_kn = self.get_spinorbit_corrections(return_wfs=True,
                                                  dft=dft,
                                                  eig_file=eig_file)
        elif eig_file is not None:
            e_kn = pickle.load(open(eig_file))[0]
        else:
            if not dft:
                try:
                    e_kn = self.gw_file['qp'][0]
                except KeyError:
                    e_kn = self.gw_file[b'qp'][0]
            else:
                e_kn = self.get_dft_eigenvalues()
        e_kn = np.sort(e_kn, axis=1)

        bandrange = self.bandrange
        ef = self.calc.get_fermi_level()
        if vac:
            evac = self.get_vacuum_level()
        else:
            evac = 0.0
        x_x, k_xc, k_x, X = self.find_k_along_path(plot_BZ=False)

        k_ibz_x = np.zeros_like(k_x)
        eGW_kn = np.zeros((len(k_x), e_kn.shape[1]))
        for n in range(e_kn.shape[1]):
            for ik in range(len(k_x)):
                ibzkpt = kd.bz2ibz_k[k_x[ik]]
                k_ibz_x[ik] = ibzkpt
                eGW_kn[ik, n] = e_kn[ibzkpt, n]

        N_occ = (eGW_kn[0] < ef).sum()
        print(N_occ, bandrange[0])
        # N_occ = int(self.calc.get_number_of_electrons()/2)
        print(' ')
        if SO:
            print('The number of Occupied bands is:', N_occ + 2 * bandrange[0])
        else:
            print('The number of Occupied bands is:', N_occ + bandrange[0])
        gap = (eGW_kn[:, N_occ].min() - eGW_kn[:, N_occ - 1].max())
        print('The bandgap is: %f' % gap)

        vbm = eGW_kn[:, N_occ - 1].max() - evac
        cbm = eGW_kn[:, N_occ].min() - evac

        if interpolate:
            xfit_k = np.linspace(x_x[0], x_x[-1], nk_Int)
            xfit_k = np.append(xfit_k, x_x)
            xfit_k = np.sort(xfit_k)
            nk_Int = len(xfit_k)
            efit_kn = np.zeros((nk_Int, eGW_kn.shape[1]))
            for n in range(eGW_kn.shape[1]):
                fit_e = InterpolatedUnivariateSpline(x_x, eGW_kn[:, n])
                efit_kn[:, n] = fit_e(xfit_k)

            results = {'x_k': xfit_k,
                       'X': X,
                       'e_kn': efit_kn - evac,
                       'ef': ef - evac,
                       'gap': gap,
                       'vbm': vbm,
                       'cbm': cbm}

            return results
        else:
            results = {'x_k': x_x,
                       'X': X,
                       'k_ibz_x': k_ibz_x,
                       'e_kn': eGW_kn - evac,
                       'ef': ef - evac,
                       'gap': gap,
                       'vbm': vbm,
                       'cbm': cbm}
            return results



"""This module defines an ELF class."""

from numpy import pi

from gpaw.fd_operators import Gradient
from gpaw.lfc import LocalizedFunctionsCollection as LFC


def _elf(nt_sg, nt_grad2_sg, taut_sg, ncut, spinpol):
    """Pseudo electron localisation function (ELF) as defined in
    Becke and Edgecombe, J. Chem. Phys., vol 92 (1990) 5397

    More comprehensive definition in
    M. Kohout and A. Savin, Int. J. Quantum Chem., vol 60 (1996) 875-882

    Arguments:
     =============== =====================================================
     ``nt_sg``       Pseudo valence density.
     ``nt_grad2_sg`` Squared norm of the density gradient.
     ``tau_sg``      Kinetic energy density.
     ``ncut``        Minimum density cutoff parameter.
     ``spinpol``     Boolean indicator for spin polarization.
     =============== =====================================================
    """

    # Fermi constant
    cF = 3.0 / 10 * (3 * pi**2)**(2.0 / 3.0)

    if spinpol:
        # Kouhut eq. (9)
        D0 = 2**(2.0 / 3.0) * cF * (nt_sg[0]**(5.0 / 3.0) +
                                    nt_sg[1]**(5.0 / 3.0))

        taut = taut_sg.sum(axis=0)
        D = taut - (nt_grad2_sg[0] / nt_sg[0] + nt_grad2_sg[1] / nt_sg[1]) / 8
    else:
        # Kouhut eq. (7)
        D0 = cF * nt_sg[0]**(5.0 / 3.0)
        taut = taut_sg[0]
        D = taut - nt_grad2_sg[0] / nt_sg[0] / 8

    elf_g = 1.0 / (1.0 + (D / D0)**2)

    if ncut is not None:
        nt = nt_sg.sum(axis=0)
        elf_g[nt < ncut] = 0.0

    return elf_g


class ELF:
    """ELF object for calculating the electronic localization function.

    Arguments:
     =============== =====================================================
     ``paw``         Instance of ``GPAW`` class.
     ``ncut``        Density cutoff below which the ELF is zero.
     =============== =====================================================
    """

    def __init__(self, paw=None, ncut=1e-6):
        """Create the ELF object."""

        if paw.wfs.mode != 'fd':
            raise NotImplementedError('Only FD mode supported for ELF')

        self.gd = paw.wfs.gd
        self.paw = paw
        self.finegd = paw.density.finegd
        self.nspins = paw.density.nspins
        self.density = paw.density

        self.ncut = ncut
        self.spinpol = (self.nspins == 2)

        self.initialize(paw)

    def initialize(self, paw):

        if not paw.initialized:
            raise RuntimeError('PAW instance is not initialized')
        paw.converge_wave_functions()

        self.tauct = LFC(self.gd,
                         [[setup.tauct] for setup in self.density.setups],
                         forces=True, cut=True)
        self.tauct.set_positions(paw.spos_ac)

        self.taut_sg = None
        self.nt_grad2_sG = self.gd.empty(self.nspins)
        self.nt_grad2_sg = None

    def interpolate(self):

        self.density.interpolate_pseudo_density()

        if self.taut_sg is None:
            self.taut_sg = self.finegd.empty(self.nspins)
            self.nt_grad2_sg = self.finegd.empty(self.nspins)

        ddr_v = [Gradient(self.finegd, v, n=3).apply for v in range(3)]
        self.nt_grad2_sg[:] = 0.0
        d_g = self.finegd.empty()

        # Transfer the densities from the coarse to the fine grid
        for s in range(self.nspins):
            self.density.interpolator.apply(self.taut_sG[s],
                                            self.taut_sg[s])
            # self.density.interpolator.apply(self.nt_grad2_sG[s],
            #                                 self.nt_grad2_sg[s])
            for v in range(3):
                ddr_v[v](self.density.nt_sg[s], d_g)
                self.nt_grad2_sg[s] += d_g**2.0

    def update(self):
        self.taut_sG = self.paw.wfs.calculate_kinetic_energy_density()

        # Add the pseudo core kinetic array
        for taut_G in self.taut_sG:
            self.tauct.add(taut_G, 1.0 / self.paw.wfs.nspins)

        # For periodic boundary conditions
        if self.paw.wfs.kd.symmetry is not None:
            self.paw.wfs.kd.symmetry.symmetrize(self.taut_sG[0],
                                                self.paw.wfs.gd)

        self.nt_grad2_sG[:] = 0.0

        d_G = self.gd.empty()

        for s in range(self.nspins):
            for v in range(3):
                Gradient(self.gd, v, n=3).apply(self.density.nt_sG[s], d_G)

                self.nt_grad2_sG[s] += d_G**2.0

        # TODO are nct from setups usable for nt_grad2_sG ?

    def get_electronic_localization_function(self, gridrefinement=1,
                                             pad=True, broadcast=True):

        # Returns dimensionless electronic localization function
        if gridrefinement == 1:
            elf_G = _elf(self.density.nt_sG, self.nt_grad2_sG,
                         self.taut_sG, self.ncut, self.spinpol)
            elf_G = self.gd.collect(elf_G, broadcast=broadcast)
            if pad:
                elf_G = self.gd.zero_pad(elf_G)
            return elf_G
        elif gridrefinement == 2:
            if self.nt_grad2_sg is None:
                self.interpolate()

            elf_g = _elf(self.density.nt_sg, self.nt_grad2_sg,
                         self.taut_sg, self.ncut, self.spinpol)
            elf_g = self.finegd.collect(elf_g, broadcast=broadcast)
            if pad:
                elf_g = self.finegd.zero_pad(elf_g)
            return elf_g
        else:
            raise NotImplementedError('Arbitrary refinement not implemented')



import numpy as np
import pytest

from gpaw.mpi import world
from gpaw.test import findpeak

from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.fxc_kernels import AdiabaticFXCCalculator
from gpaw.response.dyson import HXCScaling, HXCKernel
from gpaw.response.susceptibility import ChiFactory
from gpaw.response.pair_functions import read_pair_function


pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='too slow for world.size < 4')


@pytest.mark.kspair
@pytest.mark.response
def test_nicl2_magnetic_response(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    q_qc = [[0., 0., 0.],
            [1. / 3., 1. / 3., 0.]]
    fxc = 'ALDA'
    rshelmax = 0
    rshewmin = None
    bg_density = 0.004
    ecut = 200
    frq_w = np.linspace(-0.15, 0.075, 16)
    eta = 0.12
    zd = ComplexFrequencyDescriptor.from_array(frq_w + 1.j * eta)
    nblocks = 4

    # ---------- Script ---------- #

    # Magnetic response calculation
    context = ResponseContext()
    gs = ResponseGroundStateAdapter.from_gpw_file(gpw_files['nicl2_pw'],
                                                  context=context)
    chiks_calc = ChiKSCalculator(gs, context=context,
                                 ecut=ecut,
                                 gammacentered=True,
                                 nblocks=nblocks)

    # Calculate the magnetic response with and without a background density
    hxc_scaling = HXCScaling('fm')
    fxc_calculator = AdiabaticFXCCalculator.from_rshe_parameters(
        gs, chiks_calc.context,
        rshelmax=rshelmax,
        rshewmin=rshewmin)
    chi_factory = ChiFactory(chiks_calc, fxc_calculator)
    bgd_hxc_scaling = HXCScaling('fm')
    bgd_fxc_calculator = AdiabaticFXCCalculator.from_rshe_parameters(
        gs, chiks_calc.context,
        bg_density=bg_density,
        rshelmax=rshelmax,
        rshewmin=rshewmin)

    # Check that the pseudo-density is nonnegative
    nt_sr, gd = gs.get_pseudo_density(gridrefinement=2)
    assert np.all(nt_sr >= 0.)

    # Plot pseudo spin-density to check exponential localization into vacuum
    # if world.rank == 0:
    #     import matplotlib.pyplot as plt
    #     N_c = gd.N_c
    #     # Plot the spin-densities above one of the Cl atoms
    #     plt.plot(range(N_c[2]), nt_sr[0, 2 * N_c[0] // 3, N_c[1] // 3])
    #     plt.plot(range(N_c[2]), nt_sr[1, 2 * N_c[0] // 3, N_c[1] // 3])
    #     plt.yscale('log')
    #     plt.show()

    filestr = 'nicl2_macro_tms'
    bgd_filestr = 'nicl2_macro_tms_bgd'
    for q, q_c in enumerate(q_qc):
        chiks, chi = chi_factory('+-', q_c, zd,
                                 fxc=fxc,
                                 hxc_scaling=hxc_scaling,
                                 txt=filestr + '_q%d.txt' % q)
        chi.write_macroscopic_component(filestr + '_q%d.csv' % q)
        if q == 0:
            # Calculate kernel with background charge
            bgd_fxc_kernel = bgd_fxc_calculator(fxc, '+-', chiks.qpd)
            bgd_hxc_kernel = HXCKernel(None, bgd_fxc_kernel, bgd_hxc_scaling)
        bgd_chi = chi_factory.dyson_solver(chiks, bgd_hxc_kernel)
        bgd_chi.write_macroscopic_component(bgd_filestr + '_q%d.csv' % q)

    context.write_timer()
    world.barrier()

    # Compare new results to test values
    check_magnons(filestr, hxc_scaling,
                  test_fxcs=0.7130,
                  test_mw0=-10.3,  # meV
                  test_mw1=-40.9,  # meV
                  test_Ipeak0=0.2306,  # a.u.
                  test_Ipeak1=0.0956,  # a.u.
                  )
    check_magnons(bgd_filestr, bgd_hxc_scaling,
                  test_fxcs=1.0826,
                  test_mw0=-10.2,  # meV
                  test_mw1=-48.0,  # meV
                  test_Ipeak0=0.2321,  # a.u.
                  test_Ipeak1=0.0956,  # a.u.
                  )


def check_magnons(filestr, hxc_scaling, *,
                  test_fxcs, test_mw0, test_mw1, test_Ipeak0, test_Ipeak1):
    # Identify magnon peaks and extract kernel scaling
    w0_w, chi0_w = read_pair_function(filestr + '_q0.csv')
    w1_w, chi1_w = read_pair_function(filestr + '_q1.csv')

    wpeak0, Ipeak0 = findpeak(w0_w, -chi0_w.imag / np.pi)
    wpeak1, Ipeak1 = findpeak(w1_w, -chi1_w.imag / np.pi)
    mw0 = wpeak0 * 1e3  # meV
    mw1 = wpeak1 * 1e3  # meV

    assert hxc_scaling.lambd is not None
    fxcs = hxc_scaling.lambd

    if world.rank == 0:
        # import matplotlib.pyplot as plt
        # plt.plot(w0_w, -chi0_w.imag / np.pi)
        # plt.plot(w1_w, -chi1_w.imag / np.pi)
        # plt.show()

        print(fxcs, mw0, mw1, Ipeak0, Ipeak1)

    # Test fxc scaling
    assert fxcs == pytest.approx(test_fxcs, abs=0.005)

    # Test magnon peaks
    assert mw0 == pytest.approx(test_mw0, abs=5.0)
    assert mw1 == pytest.approx(test_mw1, abs=10.0)

    # Test peak intensities
    assert Ipeak0 == pytest.approx(test_Ipeak0, abs=0.01)
    assert Ipeak1 == pytest.approx(test_Ipeak1, abs=0.01)



import numpy as np
import pytest
from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world
from gpaw.poisson import PoissonSolver
from gpaw.test import equal


@pytest.mark.rttddft
def test_lcaotddft_fxc_rpa(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                poissonsolver=PoissonSolver('fd', eps=1e-16),
                convergence={'density': 1e-8},
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation with fxc
    td_calc = LCAOTDDFT('gs.gpw', fxc='RPA', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)
    world.barrier()

    # Check dipole moment file
    data = np.loadtxt('dm.dat')[:, 2:].ravel()
    if 0:
        from gpaw.test import print_reference
        print_reference(data, 'ref', '%.12le')

    ref = [-9.383700894739e-16,
           -9.338586948130e-16,
           2.131582675483e-14,
           8.679923327633e-15,
           7.529517689096e-15,
           2.074867751820e-14,
           1.960742702185e-05,
           1.960742702128e-05,
           1.804030746475e-05,
           3.761997205571e-05,
           3.761997205562e-05,
           3.596681520068e-05,
           5.257367158160e-05,
           5.257367158104e-05,
           5.366663490365e-05]

    tol = 1e-9
    equal(data, ref, tol)
